"""
This module contains functions for interacting with members/roles/user setup in EZOfficeInventory
"""

import logging
import os
import time

from ezoff._auth import Decorators
from ezoff._helpers import http_get, http_patch, http_post, http_put
from ezoff.data_model import CustomRole, Member, MemberCreate, Team, UserListing
from ezoff.exceptions import NoDataReturned

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def member_create(
    first_name: str | None,
    last_name: str,
    role_id: int,
    email: str,
    employee_identification_number: str | None = None,
    description: str | None = None,
    department: str | None = None,
    team_ids: list[int] | None = None,
    user_listing_id: int | None = None,
    work_location: int | None = None,
    login_enabled: bool | None = None,
    subscribed_to_emails: bool | None = None,
    skip_confirmation_email: bool | None = None,
    address_name: str | None = None,
    address: str | None = None,
    address_line_2: str | None = None,
    city: str | None = None,
    state: str | None = None,
    zip_code: str | None = None,
    country: str | None = None,
    fax: str | None = None,
    phone_number: str | None = None,
    image_url: str | None = None,
    custom_fields: list[dict] | None = None,
) -> Member | None:
    """
    Create a new member.

    :param first_name: First name of the member
    :type first_name: str | None
    :param last_name: Last name of the member
    :type last_name: str
    :param role_id: Role ID for the member, corresponds to what permissions they have
    :type role_id: int
    :param email: Email address of the member. Note: must be unique.
    :type email: str
    :param employee_identification_number: Employee ID number for the member
    :type employee_identification_number: str, optional
    :param description: Description of the member
    :type description: str, optional
    :param department: Department the member belongs to
    :type department: str, optional
    :param team_ids: List of team IDs to assign the member to
    :type team_ids: list[int], optional
    :param user_listing_id: User listing ID to associate with the member
    :type user_listing_id: int, optional
    :param login_enabled: Whether the member can log in. Non-login members can be used for non-employees or simply tracking purposes.
    :type login_enabled: bool, optional
    :param subscribed_to_emails: Whether the member is subscribed to emails
    :type subscribed_to_emails: bool, optional
    :param skip_confirmation_email: Whether to skip sending a confirmation email
    :type skip_confirmation_email: bool, optional
    :param address_name: Name associated with the member's address
    :type address_name: str, optional
    :param address: Address of the member
    :type address: str, optional
    :param address_line_2: Second line of the member's address
    :type address_line_2: str, optional
    :param city: City of the member's address
    :type city: str, optional
    :param state: State of the member's address
    :type state: str, optional
    :param zip_code: Zip code of the member's address
    :type zip_code: str, optional
    :param country: Country of the member's address
    :type country: str, optional
    :param fax: Fax number of the member
    :type fax: str, optional
    :param phone_number: Phone number of the member
    :type phone_number: str, optional
    :param image_url: URL of the member's image
    :type image_url: str, optional
    :param custom_fields: List of custom fields for the member
    :type custom_fields: list[dict], optional
    :return: The created member, or None if creation failed
    :rtype: Member | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members"
    payload = {"member": params}
    response = http_post(url=url, payload=payload, title="Member Create")

    if response.status_code == 200 and "member" in response.json():
        return Member(**response.json()["member"])
    else:
        return None


@Decorators.check_env_vars
def members_create(members: list[MemberCreate]) -> list[Member] | None:
    """
    Creates new members in bulk.

    :param members: A list of MemberCreate objects representing the members to create. Same fields present in member_create()
    :return: The list of created Members if successful, else None
    :rtype: list[Member]
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members/bulk_create"
    payload = {"members": [member.model_dump(exclude_none=True) for member in members]}
    response = http_post(url=url, payload=payload, title="Members Create")

    if response.status_code == 200 and "members" in response.json():
        return [Member(**x) for x in response.json()["members"]]
    else:
        return None


@Decorators.check_env_vars
def member_return(member_id: int) -> Member | None:
    """
    Returns a particular member.

    :param member_id: The ID of the member to return
    :return: The member if found, else None
    :rtype: Member | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members/{member_id}"
    headers = {
        "Accept": "application/json",
        "Authorization": "Bearer " + os.environ["EZO_TOKEN"],
        "Cache-Control": "no-cache",
        "Host": f"{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
    }
    response = http_post(url=url, headers=headers, title="Member Return")

    if response.status_code == 200 and "member" in response.json():
        return Member(**response.json()["member"])
    else:
        return None


@Decorators.check_env_vars
def members_return(filter: dict | None = None) -> list[Member]:
    """
    Returns all members. Optionally, filter by one or more member fields.

    :param filter:  Dictionary of member fields and the values to filter results by.
    :type filter: dict, optional
    :return: List of members
    :rtype: list[Member]
    """

    if filter:
        for field in filter:
            if field not in (
                list(Member.model_fields.keys())
                # Additional valid filters
                + [
                    "all",
                    "login_enabled",
                    "external",
                    "inactive",
                    "inactive_members_with_items",
                    "inactive_members_with_pending_associations",
                    "location_id",
                ]
            ):
                raise ValueError(f"'{field}' is not a valid field for a member.")
        filter = {"filters": filter}
    else:
        filter = None

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members"

    all_members = []
    while True:
        response = http_get(url=url, payload=filter, title="Members Return")
        data = response.json()

        if "members" not in data:
            raise NoDataReturned(f"No members found: {response.content}")

        all_members.extend(data["members"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Member(**x) for x in all_members]


@Decorators.check_env_vars
def member_update(member_id: int, update_data: dict) -> Member | None:
    """
    Updates a particular member.

    :param member_id: The ID of the member to update
    :param update_data: Dictionary of fields to update on the member. See Member model for valid fields.
    :return: The updated member if successful, else None
    :rtype: Member | None
    """

    for field in update_data:
        if field not in Member.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a member.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members/{member_id}"
    headers = ({"Authorization": "Bearer " + os.environ["EZO_TOKEN"]},)
    payload = ({"member": update_data},)
    response = http_patch(
        url=url, headers=headers, payload=payload, title="Member Update"
    )

    if response.status_code == 200 and "member" in response.json():
        return Member(**response.json()["member"])
    else:
        return None


@Decorators.check_env_vars
def member_activate(member_id: int) -> Member | None:
    """
    Activates a particular member.

    :param member_id: The ID of the member to activate
    :return: The activated member if successful, else None
    :rtype: Member | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members/{member_id}/activate"
    response = http_put(url=url, title="Member Activate")

    if response.status_code == 200 and "member" in response.json():
        return Member(**response.json()["member"])
    else:
        return None


@Decorators.check_env_vars
def member_deactivate(member_id: int) -> Member | None:
    """
    Deactivates a particular member.

    :param member_id: The ID of the member to deactivate
    :return: The deactivated member if successful, else None
    :rtype: Member | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/members/{member_id}/deactivate"
    response = http_put(url=url, title="Member Deactivate")

    if response.status_code == 200 and "member" in response.json():
        return Member(**response.json()["member"])
    else:
        return None


@Decorators.check_env_vars
def custom_roles_return() -> list[CustomRole]:
    """
    Get all custom roles.

    :return: List of custom roles
    :rtype: list[CustomRole]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/custom_roles"

    all_custom_roles = []
    while True:
        response = http_get(url=url, title="Custom Roles Return")
        data = response.json()

        if "custom_roles" not in data:
            raise NoDataReturned(f"No custom roles found: {response.content}")

        all_custom_roles.extend(data["custom_roles"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [CustomRole(**x) for x in all_custom_roles]


@Decorators.check_env_vars
def custom_role_update(custom_role_id: int, update_data) -> CustomRole | None:
    """
    Updates a particular custom role.

    :param custom_role_id: The ID of the custom role to update
    :param update_data: Dictionary of fields to update on the custom role. See CustomRole model for valid fields.
    :return: The updated custom role if successful, else None
    :rtype: CustomRole | None
    """

    for field in update_data:
        if field not in CustomRole.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a custom role.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/custom_roles/{custom_role_id}"
    payload = {"custom_role": update_data}
    response = http_patch(url=url, payload=payload, title="Custom Role Update")

    if response.status_code == 200 and "custom_role" in response.json():
        return CustomRole(**response.json()["custom_role"])
    else:
        return None


@Decorators.check_env_vars
def teams_return() -> list[Team]:
    """
    Get all teams.

    :return: List of teams
    :rtype: list[Team]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/teams"

    all_teams = []
    while True:
        response = http_get(url=url, title="Teams Return")
        data = response.json()

        if "teams" not in data:
            raise NoDataReturned(f"No teams found: {response.content}")

        all_teams.extend(data["teams"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Team(**x) for x in all_teams]


@Decorators.check_env_vars
def user_listings_return() -> list[UserListing]:
    """
    Returns all user listings.
    Note: This API endpoint is documented, but I only ever get a 403 when
    trying to use it. Even though obviously using the API key. Not sure.

    :return: List of user listings
    :rtype: list[UserListing]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/user_listings"

    all_user_listings = []
    while True:
        response = http_get(url=url, title="User Listings Return")
        data = response.json()

        if "user_listings" not in data:
            raise NoDataReturned(f"No user_listings found: {response.content}")

        all_user_listings.extend(data["user_listings"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [UserListing(**x) for x in all_user_listings]



import logging
import os
import time

from ezoff._helpers import http_get
from ezoff.data_model import RetireReason

logger = logging.getLogger(__name__)


def retire_reasons_return() -> list[RetireReason]:
    """
    Returns all retire reasons.

    :return: A list of all retire reasons.
    :rtype: list[RetireReason]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/retire_reasons"

    all_retire_reasons = []

    while True:
        response = http_get(url=url, title="Retire Reasons Return")
        data = response.json()

        if "retire_reasons" not in data:
            logger.error(f"Error, could not get retire reasons: {response.content}")
            raise Exception(f"Error, could not get retire reasons: {response.content}")

        all_retire_reasons.extend(data["retire_reasons"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [RetireReason(**x) for x in all_retire_reasons]



"""
Covers everything related to groups and subgroups in EZOfficeInventory
"""

import logging
import os
import time
from typing import Literal

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get, http_patch, http_delete
from ezoff.data_model import DepreciationRate, Group, ResponseMessages

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def group_create(
    name: str,
    description: str | None = None,
    asset_depreciation_mode: Literal["Useful Life", "Percentage"] | None = None,
    triage_completion_period: int | None = None,
    triage_completion_period_basis: (
        Literal["minutes", "hours", "days", "weeks", "months", "indefinite"] | None
    ) = None,
    allow_staff_to_set_checkout_duration: bool | None = None,
    staff_checkout_duration_months: int | None = None,
    staff_checkout_duration_weeks: int | None = None,
    staff_checkout_duration_days: int | None = None,
    staff_checkout_duration_hours: int | None = None,
    depreciation_rates: list[DepreciationRate] | None = None,
) -> Group | None:
    """
    Creates a new top-level group for items.

    :param name: Name of the group
    :type name: str
    :param description: Description of the group
    :type description: str, optional
    :param asset_depreciation_mode: The mode assets in the group will depreciate in, either 'Useful Life' or 'Percentage'
    :type asset_depreciation_mode: str, optional
    :param triage_completion_period: The time period within which the triage must be completed
    :type triage_completion_period: int, optional
    :param triage_completion_period_basis: The basis for the triage completion period, either 'minutes', 'hours', 'days', 'weeks', 'months', or 'indefinite'
    :type triage_completion_period_basis: str, optional
    :param allow_staff_to_set_checkout_duration: Whether staff are allowed to set checkout duration for assets in this group
    :type allow_staff_to_set_checkout_duration: bool, optional
    :param staff_checkout_duration_months: The number of months staff can set for checkout duration
    :type staff_checkout_duration_months: int, optional
    :param staff_checkout_duration_weeks: The number of weeks staff can set for checkout duration
    :type staff_checkout_duration_weeks: int, optional
    :param staff_checkout_duration_days: The number of days staff can set for checkout duration
    :type staff_checkout_duration_days: int, optional
    :param staff_checkout_duration_hours: The number of hours staff can set for checkout duration
    :type staff_checkout_duration_hours: int, optional
    :param depreciation_rates: A list of depreciation rates to apply to the group
    :type depreciation_rates: list of DepreciationRate, optional
    :return: The created group, or None if the creation failed
    :rtype: Group or None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups"
    payload = {"group": params}
    response = http_post(url=url, payload=payload, title="Group Create")

    if response.status_code == 200 and "group" in response.json(0):
        return Group(**response.json()["group"])
    else:
        return None


@Decorators.check_env_vars
def group_return(group_id: int) -> Group | None:
    """
    Returns a particular group.

    :param group_id: The ID of the group to return
    :type group_id: int
    :return: The group with the specified ID, or None if not found
    :rtype: Group or None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}"
    response = http_get(url=url, title="Group Return")

    if response.status_code == 200 and "group" in response.json():
        return Group(**response.json()["group"])
    else:
        return None


@Decorators.check_env_vars
def groups_return() -> list[Group]:
    """
    Returns all groups.

    :return: A list of all groups
    :rtype: list of Group
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups"

    all_groups = []

    while True:
        response = http_get(url=url, title="Groups Return")
        data = response.json()

        if "groups" not in data:
            logger.error(f"Error, could not get groups: {response.content}")
            raise Exception(f"Error, could not get groups: {response.content}")

        all_groups.extend(data["groups"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Group(**x) for x in all_groups]


@Decorators.check_env_vars
def group_update(group_id: int, update_data: dict) -> Group | None:
    """
    Updates a particular group.

    :param group_id: The ID of the group to update
    :type group_id: int
    :param update_data: A dictionary of fields to update and their new values
    :type update_data: dict
    :return: The updated group, or None if the update failed
    :rtype: Group | None
    """

    for field in update_data:
        if field not in Group.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a group.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}"
    payload = {"group": update_data}
    response = http_patch(url=url, payload=payload, title="Group Update")

    if response.status_code == 200 and "group" in response.json():
        return Group(**response.json()["group"])
    else:
        return None


@Decorators.check_env_vars
def group_delete(group_id: int) -> ResponseMessages | None:
    """
    Deletes a particular group.

    :param group_id: The ID of the group to delete
    :type group_id: int
    :return: ResponseMessages object if there are any messages, else None.
    :rtype: ResponseMessages | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}"
    response = http_delete(url=url, title="Group Delete")

    if response.status_code == 200 and "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None


@Decorators.check_env_vars
def subgroup_create(
    parent_id: int,
    name: str,
    description: str | None = None,
    asset_depreciation_mode: Literal["Useful Life", "Percentage"] | None = None,
    triage_completion_period: int | None = None,
    triage_completion_period_basis: (
        Literal["minutes", "hours", "days", "weeks", "months", "indefinite"] | None
    ) = None,
    allow_staff_to_set_checkout_duration: bool | None = None,
    staff_checkout_duration_months: int | None = None,
    staff_checkout_duration_weeks: int | None = None,
    staff_checkout_duration_days: int | None = None,
    staff_checkout_duration_hours: int | None = None,
    depreciation_rates: list[DepreciationRate] | None = None,
) -> Group | None:
    """
    Creates a subgroup under a parent group.

    :param parent_id: The ID of the parent group under which to create the subgroup
    :type parent_id: int
    :param name: Name of the subgroup
    :type name: str
    :param description: Description of the subgroup
    :type description: str, optional
    :param asset_depreciation_mode: The mode assets in the subgroup will depreciate in, either 'Useful Life' or 'Percentage'
    :type asset_depreciation_mode: str, optional
    :param triage_completion_period: The time period within which the triage must be completed
    :type triage_completion_period: int, optional
    :param triage_completion_period_basis: The basis for the triage completion period, either 'minutes', 'hours', 'days', 'weeks', 'months', or 'indefinite'
    :type triage_completion_period_basis: str, optional
    :param allow_staff_to_set_checkout_duration: Whether staff are allowed to set checkout duration for assets in this subgroup
    :type allow_staff_to_set_checkout_duration: bool, optional
    :param staff_checkout_duration_months: The number of months staff can set for checkout duration
    :type staff_checkout_duration_months: int, optional
    :param staff_checkout_duration_weeks: The number of weeks staff can set for checkout duration
    :type staff_checkout_duration_weeks: int, optional
    :param staff_checkout_duration_days: The number of days staff can set for checkout duration
    :type staff_checkout_duration_days: int, optional
    :param staff_checkout_duration_hours: The number of hours staff can set for checkout duration
    :type staff_checkout_duration_hours: int, optional
    :param depreciation_rates: A list of depreciation rates to apply to the subgroup
    :type depreciation_rates: list of DepreciationRate, optional
    :return: The created subgroup, or None if the creation failed
    :rtype: Group | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{parent_id}/sub_groups"
    response = http_post(url=url, payload=params, title="SubGroup Create")

    if response.status_code == 200 and "sub_group" in response.json():
        return Group(**response.json()["sub_group"])
    else:
        return None


@Decorators.check_env_vars
def subgroup_return(group_id: int, subgroup_id: int) -> Group | None:
    """
    Returns a particular subgroup.

    :param group_id: The ID of the parent group
    :type group_id: int
    :param subgroup_id: The ID of the subgroup to return
    :type subgroup_id: int
    :return: The subgroup with the specified ID, or None if not found
    :rtype: Group | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}/sub_groups/{subgroup_id}"
    response = http_get(url=url, title="SubGroup Return")

    if response.status_code == 200 and "sub_group" in response.json():
        return Group(**response.json()["sub_group"])
    else:
        return None


@Decorators.check_env_vars
def subgroups_return(group_id: int) -> list[Group]:
    """
    Get all subgroups under a particular group.

    :param group_id: Filter to get subgroups of a specific group
    :type group_id: int
    :return: A list of all subgroups under the specified group
    :rtype: list[Group]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}"

    all_subgroups = []

    while True:
        response = http_get(url=url, title="SubGroups Return")
        data = response.json()

        if "groups" not in data:
            logger.error(f"Error, could not get subgroups: {response.content}")
            raise Exception(f"Error, could not get subgroups: {response.content}")

        all_subgroups.extend(data["sub_groups"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Group(**x) for x in all_subgroups]


@Decorators.check_env_vars
def subgroup_update(group_id: int, subgroup_id: int, update_data: dict) -> Group | None:
    """
    Updates a particular subgroup.

    :param group_id: The ID of the parent group
    :type group_id: int
    :param subgroup_id: The ID of the subgroup to update
    :type subgroup_id: int
    :param update_data: A dictionary of fields to update and their new values
    :type update_data: dict
    :return: The updated subgroup, or None if the update failed
    :rtype: Group | None
    """
    for field in update_data:
        if field not in Group.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a group.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}/sub_groups/{subgroup_id}"
    response = http_patch(url=url, payload=update_data, title="SubGroup Update")

    if response.status_code == 200 and "sub_group" in response.json():
        return Group(**response.json()["sub_group"])
    else:
        return None


@Decorators.check_env_vars
def subgroup_delete(group_id: int, subgroup_id: int) -> ResponseMessages | None:
    """
    Deletes a particular subgroup.

    :param group_id: The ID of the parent group
    :type group_id: int
    :param subgroup_id: The ID of the subgroup to delete
    :type subgroup_id: int
    :return: ResponseMessages object if there are any messages, else None.
    :rtype: ResponseMessages | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/groups/{group_id}/sub_groups/{subgroup_id}"
    response = http_delete(url=url, title="SubGroup Delete")

    if response.status_code == 200 and "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None


# TODO Add group depreciation rates



"""
Module contains any custom exceptions defined for use in package.
"""


class AssetException(Exception):
    """
    Parent class for exceptions related to EZ-Office Fixed Assets.
    """

    def __init__(self, asset_id: str):
        super().__init__("")
        self.asset_id = asset_id


class AssetNotFound(AssetException):
    def __str__(self):
        return f"Asset {self.asset_id} not found in EZ-Office."


class AssetDuplicateIdentificationNumber(Exception):
    pass


class ChecklistLinkError(Exception):
    pass


class ChecklistNotFound(Exception):
    pass


class LocationNotFound(Exception):
    pass


class MemberNotFound(Exception):
    pass


class NoDataReturned(Exception):
    pass


class WorkOrderCompleted(Exception):
    pass


class WorkOrderNotFound(Exception):
    pass


class WorkOrderUpdateError(Exception):
    pass



import logging
import os
import time
from datetime import datetime

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_put, http_get
from ezoff.data_model import Package, ResponseMessages

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def package_create(
    name: str,
    description: str | None = None,
    asset_ids: list[int] | None = None,
    arbitration: str | None = None,
) -> Package | None:
    """
    Create a new asset package.

    :param name: Name of the package
    :type name: str
    :param description: Description of the package
    :type description: str, optional
    :param asset_ids: List of asset IDs to include in the package
    :type asset_ids: list[int], optional
    :param arbitration: Arbitration details for the package
    :type arbitration: str, optional
    :return: The created package if successful, else None
    :rtype: Package | None
    """
    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/packages"
    response = http_post(url=url, payload={"package": params}, title="Package Create")

    if response.status_code == 200 and "package" in response.json():
        return Package(**response.json()["package"])
    else:
        return None


@Decorators.check_env_vars
def package_return(package_id: int) -> Package | None:
    """
    Returns a particular package.

    :param package_id: The ID of the package to retrieve
    :return: The package if found, else None
    :rtype: Package | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/packages/{package_id}"
    response = http_get(url=url, title="Package Return")

    if response.status_code == 200 and "package" in response.json():
        return Package(**response.json()["package"])
    else:
        return None


@Decorators.check_env_vars
def packages_return() -> list[Package]:
    """
    Returns all packages.

    :return: List of all packages
    :rtype: list[Package]
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/packages"

    all_packages = []

    while True:
        response = http_get(url=url, title="Packages Return")
        data = response.json()

        if "packages" not in data:
            logger.error(f"Error, could not get packages: {response.content}")
            raise Exception(f"Error, could not get packages: {response.content}")

        all_packages.extend(data["packages"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Package(**x) for x in all_packages]


@Decorators.check_env_vars
def package_checkin(
    package_id: int, comments: str, location_id: int, checkin_date: datetime
) -> ResponseMessages | None:
    """
    Checks in an asset package.

    :param package_id: The ID of the package to check in
    :type package_id: int
    :param comments: Comments regarding the check-in
    :type comments: str
    :param location_id: The ID of the location where the package is being checked in
    :type location_id: int
    :param checkin_date: The date and time of the check-in
    :type checkin_date: datetime
    :return: Response messages if successful, else None
    :rtype: ResponseMessages | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/packages/{package_id}/checkin"
    response = http_put(
        url=url,
        payload={
            "package": {
                "comments": comments,
                "location_id": location_id,
                "checkin_date": checkin_date,
            }
        },
        title="Package Checkin",
    )

    if response.status_code == 200 and "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None



"""
This module contains functions for interacting with members/roles/user setup in EZOfficeInventory
"""

import logging
import os
import time
from typing import Optional

from ezoff._auth import Decorators
from ezoff._helpers import http_get, http_patch

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def members_return_v1(filter: Optional[dict]) -> list[dict]:
    """
    Get members from EZOfficeInventory
    Optionally filter by email, employee_identification_number, or status
    https://ezo.io/ezofficeinventory/developers/#api-retrieve-members

    :param filter: Dictionary of filter parameters
    :type filter: Optional[dict]
    :return: List of members
    :rtype: list[dict]
    """

    if filter is not None:
        if "filter" not in filter or "filter_val" not in filter:
            raise ValueError("filter must have 'filter' and 'filter_val' keys")

        if filter["filter"] not in [
            "email",
            "employee_identification_number",
            "status",
        ]:
            raise ValueError(
                "filter['filter'] must be one of 'email', 'employee_identification_number', 'status'"
            )

    # url = os.environ["EZO_BASE_URL"] + "members.api"
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/members.api"

    page = 1
    all_members = []

    while True:
        params = {"page": page, "include_custom_fields": "true"}
        if filter is not None:
            params.update(filter)

        response = http_get(url=url, params=params, title="Members Return v1")
        data = response.json()

        if "members" not in data:
            logger.error(f"Error, could not get members: {data}")
            raise Exception(f"Error, could not get members: {response.content}")

        all_members.extend(data["members"])

        if "total_pages" not in data:
            break

        if page >= data["total_pages"]:
            break

        page += 1

        # Potentially running into rate limiting issues with this endpoint
        # Sleep for a second to avoid this
        time.sleep(1)

    return all_members


@Decorators.check_env_vars
def member_update_v1(member_id: int, member: dict) -> dict:
    """
    Update a member with v1 API. Re-added this as the v2 member update endpoint
    doesn't yet support changing the location ID field.

    Note: If updating a customer that has an email, you should include the email
    in the member dict. If you don't, it will get removed for some reason. Not sure
    why as I'm using PATCH. So presumably should only be touching the keys that
    are specified.
    https://ezo.io/ezofficeinventory/developers/#api-update-member

    :param member_id: The ID of the member to update
    :type member_id: int
    :param member: Dictionary of member data to update.
    :type member: dict
    :return: The updated member data
    :rtype: dict
    """

    # Remove any keys that are not valid
    valid_keys = [
        "user[email]",
        "user[employee_id]",
        "user[employee_identification_number]",
        "user[role_id]",
        "user[team_id]",
        "user[user_listing_id]",
        "user[first_name]",
        "user[last_name]",
        "user[address_name]",
        "user[address]",
        "user[address_line_2]",
        "user[city]",
        "user[state]",
        "user[country]",
        "user[phone_number]",
        "user[fax]",
        "user[login_enabled]",
        "user[subscribed_to_emails]",
        "user[display_picture]",
        "user[unsubscribed_by_id]",
        "user[authorization_amount]",
        "user[vendor_id]",
        "user[time_zone]",
        "user[hourly_rate]",
        "user[offboarding_date]",
        "user[location_id]",
        "user[default_address_id]",
        "user[description]",
        "user[department]",
        "skip_confirmation_email",
    ]

    # Check for custom attributes
    member = {
        k: v
        for k, v in member.items()
        if k in valid_keys or k.startswith("user[custom_attributes]")
    }

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/members/{member_id}.api"
    # member was being passed to data param of requests.patch. might need to address.
    response = http_patch(url=url, json=member, title="Member Update v1")

    return response.json()



"""
Module for interacting with the EZO API

Developer Documentation
API v1: https://ezo.io/ezofficeinventory/developers/
API v2: https://www.ezofficeinventory.com/api-docs/index.html
"""

from .assets import (
    asset_activate,
    asset_checkin,
    asset_checkout,
    asset_create,
    asset_delete,
    asset_history_return,
    asset_public_link_return,
    asset_retire,
    asset_return,
    asset_update,
    asset_verification_request,
    assets_return,
    assets_search,
    assets_token_input_return,
)
from .bundle import bundle_create, bundle_return, bundles_return
from .checklists import checklists_return
from .data_model import (
    Asset,
    AssetHistoryItem,
    Bundle,
    Checklist,
    ChecklistLineItem,
    Component,
    CustomFieldHistoryItem,
    CustomRole,
    DepreciationRate,
    Group,
    Inventory,
    LinkedInventory,
    Location,
    Member,
    MemberCreate,
    Package,
    Project,
    PurchaseOrder,
    Reservation,
    ResponseMessages,
    RetireReason,
    StockAsset,
    StockHistoryItem,
    Team,
    TokenInput,
    UserListing,
    Vendor,
    WorkLog,
    WorkOrder,
)
from .enums import AssetClass, CustomFieldID, LocationClass, RentLoan, ResourceType
from .exceptions import (
    AssetException,
    AssetNotFound,
    AssetDuplicateIdentificationNumber,
    ChecklistLinkError,
    ChecklistNotFound,
    LocationNotFound,
    MemberNotFound,
    NoDataReturned,
    WorkOrderCompleted,
    WorkOrderNotFound,
    WorkOrderUpdateError,
)
from .ezo_cache import AssetCache, EzoCache, LocationCache, MemberCache, WorkOrderCache
from .groups import (
    group_create,
    group_delete,
    group_return,
    group_update,
    groups_return,
    subgroup_create,
    subgroup_delete,
    subgroup_return,
    subgroup_update,
    subgroups_return,
)
from .inventory import (
    inventories_return,
    inventories_search,
    inventory_activate,
    inventory_add_stock,
    inventory_create,
    inventory_custom_field_history_return,
    inventory_delete,
    inventory_history_return,
    inventory_link_to_project,
    inventory_quantity_by_location_return,
    inventory_remove_stock,
    inventory_reservations_return,
    inventory_retire,
    inventory_return,
    inventory_transfer_stock,
    inventory_unlink_from_project,
    inventory_update_location,
)
from .locations import (
    location_activate,
    location_create,
    location_deactivate,
    location_return,
    location_update,
    locations_return,
)
from .members import (
    custom_role_update,
    custom_roles_return,
    member_activate,
    member_create,
    member_deactivate,
    member_return,
    member_update,
    members_create,
    members_return,
    teams_return,
    user_listings_return,
)
from .members_v1 import member_update_v1, members_return_v1
from .packages import package_checkin, package_create, package_return, packages_return
from .projects import (
    project_create,
    project_mark_complete,
    project_mark_in_progress,
    project_return,
    projects_return,
)
from .purchase_orders import (
    purchase_order_create,
    purchase_order_return,
    purchase_orders_return,
)
from .retire_reasons import retire_reasons_return
from .stock_assets import (
    stock_asset_create,
    stock_asset_delete,
    stock_asset_return,
    stock_assets_return,
    stock_assets_search,
)
from .vendors import vendor_create, vendor_return, vendor_update, vendors_return
from .work_orders import (
    service_create,
    work_order_add_checklist,
    work_order_add_component,
    work_order_add_linked_inv,
    work_order_add_linked_po,
    work_order_add_linked_wo,
    work_order_add_work_log,
    work_order_create,
    work_order_delete,
    work_order_force_complete,
    work_order_linked_inventory_return,
    work_order_linked_work_orders_return,
    work_order_mark_complete,
    work_order_mark_in_progress,
    work_order_mark_on_hold,
    work_order_remove_checklist,
    work_order_remove_linked_po,
    work_order_remove_linked_wo,
    work_order_return,
    work_order_routing_update,
    work_order_types_return,
    work_order_update,
    work_order_update_checklist,
    work_order_work_logs_return,
    work_orders_delete,
    work_orders_end_component_service,
    work_orders_return,
    work_orders_search,
    work_orders_start_component_service,
)

__all__ = [
    "asset_activate",
    "asset_checkin",
    "asset_checkout",
    "asset_create",
    "asset_delete",
    "asset_history_return",
    "asset_public_link_return",
    "asset_retire",
    "asset_return",
    "asset_update",
    "asset_verification_request",
    "AssetCache",
    "AssetClass",
    "assets_return",
    "assets_search",
    "assets_token_input_return",
    "bundle_create",
    "bundle_return",
    "bundles_return",
    "checklists_return",
    "custom_role_update",
    "custom_roles_return",
    "CustomFieldID",
    "EzoCache",
    "group_create",
    "group_delete",
    "group_return",
    "group_update",
    "groups_return",
    "inventories_return",
    "inventories_search",
    "inventory_activate",
    "inventory_add_stock",
    "inventory_create",
    "inventory_custom_field_history_return",
    "inventory_delete",
    "inventory_history_return",
    "inventory_link_to_project",
    "inventory_quantity_by_location_return",
    "inventory_remove_stock",
    "inventory_reservations_return",
    "inventory_retire",
    "inventory_return",
    "inventory_transfer_stock",
    "inventory_unlink_from_project",
    "inventory_update_location",
    "location_activate",
    "location_create",
    "location_deactivate",
    "location_return",
    "location_update",
    "LocationCache",
    "LocationClass",
    "locations_return",
    "member_activate",
    "member_create",
    "member_deactivate",
    "member_return",
    "members_return_v1",
    "member_update_v1",
    "member_update",
    "MemberCache",
    "members_create",
    "members_return",
    "package_checkin",
    "package_create",
    "package_return",
    "packages_return",
    "project_create",
    "project_mark_complete",
    "project_mark_in_progress",
    "project_return",
    "projects_return",
    "purchase_order_create",
    "purchase_order_return",
    "purchase_orders_return",
    "RentLoan",
    "ResourceType",
    "retire_reasons_return",
    "service_create",
    "stock_asset_create",
    "stock_asset_delete",
    "stock_asset_return",
    "stock_assets_return",
    "stock_assets_search",
    "subgroup_create",
    "subgroup_delete",
    "subgroup_return",
    "subgroup_update",
    "subgroups_return",
    "teams_return",
    "user_listings_return",
    "vendor_create",
    "vendor_return",
    "vendor_update",
    "vendors_return",
    "work_order_add_checklist",
    "work_order_add_component",
    "work_order_add_linked_inv",
    "work_order_add_linked_po",
    "work_order_add_linked_wo",
    "work_order_add_work_log",
    "work_order_create",
    "work_order_delete",
    "work_order_linked_inventory_return",
    "work_order_linked_work_orders_return",
    "work_order_force_complete",
    "work_order_mark_complete",
    "work_order_mark_in_progress",
    "work_order_mark_on_hold",
    "work_order_remove_checklist",
    "work_order_remove_linked_po",
    "work_order_remove_linked_wo",
    "work_order_return",
    "work_order_routing_update",
    "work_order_types_return",
    "work_order_update_checklist",
    "work_order_update",
    "work_order_work_logs_return",
    "work_orders_delete",
    "work_orders_end_component_service",
    "work_orders_return",
    "work_orders_search",
    "work_orders_start_component_service",
    "WorkOrderCache",
    "ResponseMessages",
    "Asset",
    "Inventory",
    "ChecklistLineItem",
    "Checklist",
    "Component",
    "Location",
    "Member",
    "MemberCreate",
    "CustomRole",
    "UserListing",
    "WorkOrder",
    "WorkLog",
    "DepreciationRate",
    "Group",
    "Team",
    "Project",
    "CustomFieldHistoryItem",
    "StockHistoryItem",
    "Reservation",
    "TokenInput",
    "AssetHistoryItem",
    "Bundle",
    "PurchaseOrder",
    "Package",
    "RetireReason",
    "Vendor",
    "StockAsset",
    "LinkedInventory",
]



import logging
import os
import time

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get
from ezoff.data_model import PurchaseOrder

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def purchase_order_create(title: str, vendor_id: int) -> PurchaseOrder | None:
    """
    Creates a new purchase order.

    :param title: Title of the purchase order.
    :type title: str
    :param vendor_id: ID of the vendor for the purchase order.
    :type vendor_id: int
    :return: The created purchase order or None if creation failed.
    :rtype: PurchaseOrder | None
    """
    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/purchase_orders"
    response = http_post(
        url=url, payload={"purchase_order": params}, title="Purchase Order Create"
    )

    if response.status_code == 200 and "purchase_order" in response.json():
        return PurchaseOrder(**response.json()["purchase_order"])
    else:
        return None


@Decorators.check_env_vars
def purchase_order_return(purchase_order_id: int) -> PurchaseOrder | None:
    """
    Returns a particular purchase order.

    :param purchase_order_id: ID of the purchase order to return.
    :type purchase_order_id: int
    :return: The requested purchase order or None if not found.
    :rtype: PurchaseOrder | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/purchase_orders/{purchase_order_id}"
    response = http_get(url=url, title="Purchase Order Return")

    if response.status_code == 200 and "purchase_order" in response.json():
        return PurchaseOrder(**response.json()["purchase_order"])
    else:
        return None


@Decorators.check_env_vars
def purchase_orders_return() -> list[PurchaseOrder]:
    """
    Returns all purchase orders.

    :return: A list of all purchase orders.
    :rtype: list[PurchaseOrder]
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/purchase_orders"

    all_purchase_orders = []

    while True:
        response = http_get(url=url, title="Purchase Orders Return")
        data = response.json()

        if "purchase_orders" not in data:
            logger.error(f"Error, could not get purchase orders: {response.content}")
            raise Exception(f"Error, could not get purchase orders: {response.content}")

        all_purchase_orders.extend(data["purchase_orders"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [PurchaseOrder(**x) for x in all_purchase_orders]


# TODO Update

# TODO Mark void

# TODO Add items

# TODO Receive items

# TODO Mark Confirmed

# TODO Add items

# TODO Delete



"""
This file contains classes for making cached calls to EZ Office API endpoints.
The parent class EzoCache contains basic caching functionality.
Child classes extend EzoCache and add endpoint specific methods.
"""

import logging
import pickle

from ezoff.assets import asset_return, assets_return
from ezoff.data_model import Asset, Location, Member, WorkOrder
from ezoff.exceptions import (
    AssetNotFound,
    LocationNotFound,
    MemberNotFound,
    WorkOrderNotFound,
)
from ezoff.locations import location_return, locations_return
from ezoff.members import member_return, members_return
from ezoff.work_orders import work_order_return, work_orders_return
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class EzoCache:
    """
    Parent class for caching EZ Office API data.

    :ivar cache: Local cache mapping entry IDs to Pydantic models.
    :ivar cache_id_nums:
    :ivar _debug: Debug mode
    :ivar _use_saved: Whether to use saved pickle files, as opposed to the API.
    :ivar _pickle_file_name: The name of the pickle file to use.
    :ivar _api_call_single: The function to use to retrieve a particular item (using an entry ID).
    :ivar _api_call_multi: The function to use to retrieve multiple items (getting all or optionally using a filter).
    :ivar _data_model: The pydantic model corresponding to the cache class.
    :ivar _not_found_exception: The exception that will be raised if no corresponding entry is found with the API.
    """

    def __init__(self, debug: bool = False, use_saved: bool = False):
        self.cache: dict[int, BaseModel] = {}
        self._debug = debug
        self._use_saved = use_saved
        self._pickle_file_name: str | None = None
        self._api_call_single = None
        self._api_call_single_param_name = "entry_id"
        self._api_call_multi = None
        self._data_model: type[BaseModel] | None = None
        self._not_found_exception: type[Exception] | None = None

    def _get_cache_entry(self, entry_id: int, force_api: bool = False) -> BaseModel:
        """
        Returns BaseModel object representing the entry identified by entry_id.
        Subsequent calls referencing the same entry_id will be retrieved from
        the local cache instead of making further calls to the EZO API.

        :param entry_id: ID of entry to return.
        :param force_api: Wether to get data from the API even if a cached copy exists.
        :raises _not_found_exception: Raised when entry_id is not found.
        :returns BaseModel: Pydantic object.
        """
        if force_api or entry_id not in self.cache:
            params = {self._api_call_single_param_name: entry_id}
            try:
                assert self._api_call_single is not None
                self.cache[entry_id] = self._api_call_single(**params)
                return self.cache[entry_id]

            except self._not_found_exception as e:
                raise self._not_found_exception(
                    f"Entry ID {entry_id} not found. {str(e)}"
                )

        return self.cache[entry_id]

    def clear(self):
        """
        Clears EZO cached data.
        """
        self.cache = {}

    def download(self, filter: dict | None = None) -> None:
        """
        Downloads EZO data into local cache.
        New data is appended to or overwrites locally cached data.

        :param filter: Body/payload filter data for limiting results. See EZ Office API v2 for filter schema.
        """
        logger.info(f"Downloading from EZ Office. Filter: {filter}")

        # Use saved pickle or save a pickle when running in debug mode.
        if self._debug:
            assert self._pickle_file_name is not None
            if self._use_saved:
                with open(self._pickle_file_name, "rb") as f:
                    cache = pickle.load(f)

            else:
                assert self._api_call_multi is not None
                cache = self._api_call_multi(filter=filter)
                with open(self._pickle_file_name, "wb") as f:
                    pickle.dump(cache, f)

        # Call EZO API if not running in debug mode.
        else:
            assert self._api_call_multi is not None
            cache = self._api_call_multi(filter=filter)

        logger.info(f"Returned {len(cache)} results.")
        self.cache = {**self.cache, **{x.id: x for x in cache}}


class AssetCache(EzoCache):
    def __init__(self, debug=False, use_saved=False):
        super().__init__(debug, use_saved)
        self.cache: dict[int, Asset] = {}
        self._pickle_file_name = "ezo_asset_cache.pkl"
        self._api_call_single = asset_return
        self._api_call_single_param_name = "asset_id"
        self._api_call_multi = assets_return
        self._data_model = Asset
        self._not_found_exception = AssetNotFound

    def asset(self, asset_id: int, force_api: bool = False):
        return self._get_cache_entry(entry_id=asset_id, force_api=force_api)

    @property
    def assets(self) -> dict[int, Asset]:
        return self.cache


class LocationCache(EzoCache):
    def __init__(self, debug=False, use_saved=False):
        super().__init__(debug, use_saved)
        self.cache: dict[int, Location] = {}
        self._pickle_file_name = "ezo_location_cache.pkl"
        self._api_call_single = location_return
        self._api_call_single_param_name = "location_id"
        self._api_call_multi = locations_return
        self._data_model = Location
        self._not_found_exception = LocationNotFound

    def location(self, location_id: int, force_api: bool = False):
        return self._get_cache_entry(entry_id=location_id, force_api=force_api)

    @property
    def locations(self) -> dict[int, Location]:
        return self.cache


class MemberCache(EzoCache):
    def __init__(self, debug=False, use_saved=False):
        super().__init__(debug, use_saved)
        self.cache: dict[int, Member] = {}
        self._pickle_file_name = "ezo_member_cache.pkl"
        self._api_call_single = member_return
        self._api_call_single_param_name = "member_id"
        self._api_call_multi = members_return
        self._data_model = Member
        self._not_found_exception = MemberNotFound

    def member(self, member_id: int, force_api: bool = False):
        return self._get_cache_entry(entry_id=member_id, force_api=force_api)

    @property
    def members(self) -> dict[int, Member]:
        return self.cache


class WorkOrderCache(EzoCache):
    def __init__(self, debug=False, use_saved=False):
        super().__init__(debug, use_saved)
        self.cache: dict[int, WorkOrder] = {}
        self._pickle_file_name = "ezo_workorder_cache.pkl"
        self._api_call_single = work_order_return
        self._api_call_single_param_name = "work_order_id"
        self._api_call_multi = work_orders_return
        self._data_model = WorkOrder
        self._not_found_exception = WorkOrderNotFound

    def work_order(self, work_order_id: int, force_api: bool = False):
        return self._get_cache_entry(entry_id=work_order_id, force_api=force_api)

    @property
    def work_orders(self) -> dict[int, WorkOrder]:
        return self.cache



"""
Covers everything related to fixed assets in EZOffice
"""

import logging
import os
import time
from datetime import date, datetime

from ezoff._auth import Decorators
from ezoff._helpers import (
    http_post,
    http_put,
    http_get,
    http_patch,
    http_delete,
)
from ezoff.data_model import Asset, AssetHistoryItem, ResponseMessages, TokenInput
from ezoff.exceptions import (
    NoDataReturned,
)

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def asset_create(
    name: str,
    group_id: int,
    location_id: int,
    sub_group_id: int | None = None,
    purchased_on: datetime | None = None,
    display_image: str | None = None,
    identifier: str | None = None,
    description: str | None = None,
    vendor_id: int | None = None,
    product_model_number: str | None = None,
    cost_price: float | None = None,
    salvage_value: float | None = None,
    arbitration: int | None = None,
    custom_fields: list[dict] | None = None,
) -> Asset | None:
    """
    Creates a new asset.

    :param name: The name of the new asset.
    :type name: str
    :param group_id: The top-level grouping to place the asset in.
    :type group_id: int
    :param location_id: The location ID to place the asset in.
    :type location_id: int
    :param sub_group_id: The sub-grouping to place the asset in.
    :type sub_group_id: int, optional
    :param purchased_on: The date the asset was purchased.
    :type purchased_on: datetime, optional
    :param display_image: URL to an image to use as the asset's display image.
    :type display_image: str, optional
    :param identifier: The asset's identifier, such as a serial number or asset tag.
    :type identifier: str, optional
    :param description: A description of the asset.
    :type description: str, optional
    :param vendor_id: The ID of the vendor the asset was purchased from.
    :type vendor_id: int, optional
    :param product_model_number: The product model number of the asset.
    :type product_model_number: str, optional
    :param cost_price: The cost price of the asset.
    :type cost_price: float, optional
    :param salvage_value: The salvage value of the asset.
    :type salvage_value: float, optional
    :param arbitration: The arbitration period of the asset in months.
    :type arbitration: int, optional
    :param custom_fields: List of custom fields to set on the asset. Each item in
        the list should be a dictionary with 'id' and 'value' keys.
    :type custom_fields: list[dict], optional
    :return: The created asset object if successful, else None.
    :rtype: Asset | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets"
    response = http_post(url=url, payload={"asset": params}, title="Asset Create")

    if response.status_code == 200 and "asset" in response.json():
        return Asset(**response.json()["asset"])
    else:
        return None


# TODO Bulk create

# TODO Reservation create


@Decorators.check_env_vars
def asset_return(asset_id: int) -> Asset | None:
    """
    Returns a particular asset.

    :param asset_id: The ID of the asset to return.
    :type asset_id: int
    :return: The asset object if found, else None.
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}"
    response = http_get(url=url, title="Asset Return")

    if response.status_code == 200 and "asset" in response.json():
        return Asset(**response.json()["asset"])
    else:
        return None


@Decorators.check_env_vars
def assets_return(filter: dict | None = None) -> list[Asset]:
    """
    Get a list of all assets, or use the filter parameter to filter results.

    :param filter:  Dictionary of asset fields and the values to filter results by.
    :type filter: dict, optional
    :return: List of assets
    :rtype: list[Asset]
    """

    if filter:
        for field in filter:
            if field not in Asset.model_fields:
                raise ValueError(f"'{field}' is not a valid field for an asset.")
        filter = {"filters": filter}
    else:
        filter = None

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets"
    headers = {
        "Accept": "application/json",
        "Authorization": "Bearer " + os.environ["EZO_TOKEN"],
        "Cache-Control": "no-cache",
        "Host": f"{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com",
        "Accept-Encoding": "gzip, deflate, br",
        "Connection": "keep-alive",
        "Content-Type": "application/json",
    }
    all_assets = []

    while True:
        response = http_get(
            url=url, payload=filter, headers=headers, title="Assets Return"
        )
        data = response.json()

        if "assets" not in data:
            raise NoDataReturned(f"No fixed assets found: {response.content}")

        all_assets.extend(data["assets"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

    return [Asset(**x) for x in all_assets]


@Decorators.check_env_vars
def assets_search(search_term: str) -> list[Asset]:
    """
    Search for assets.
    The equivalent of the search bar in the EZO UI.
    May not return all assets that match the search term. Better to use the
    assets_return function with filters if you know what you're looking for.

    :param search_term: The search term to search assets by.
    :type search_term: str
    :return: List of Asset objects matching the search term.
    :rtype: list[Asset]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/search"
    all_assets = []

    while True:
        response = http_get(
            url=url, params={"search": search_term}, title="Assets Search"
        )
        data = response.json()

        if "assets" not in data:
            logger.error(f"Error, could not get assets: {response.content}")
            raise Exception(f"Error, could not get assets: {response.content}")

        all_assets.extend(data["assets"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Asset(**x) for x in all_assets]


@Decorators.check_env_vars
def asset_public_link_return(asset_id: int) -> str | None:
    """
    Returns the public link for a particular asset.

    Note: not sure if might get more than one link for a given asset. The API
    endpoint 'get_public_links' implies so, but in my testing I've only gotten an individual link
    for a given asset.

    :param asset_id: The ID of the asset to get the public link for.
    :type asset_id: int
    :return: The public link if found, else None.
    :rtype: str | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/get_public_links"
    response = http_get(url=url, title="Asset Public Link Return")

    if "link" in response.json():
        return response.json()["link"]
    else:
        return None


# TODO Booked dates return

# TODO Reservations return


@Decorators.check_env_vars
def assets_token_input_return(q: str) -> list[TokenInput]:
    """
    This isn't an official endpoint in the EZO API. It's used to populate
    the token input dropdowns in the EZO UI. Was previously used in the
    to get the ID needed to filter work orders by asset, but the v2 API
    support directly filtering work orders by asset ID now. Keeping it here
    for posterity.

    Note: If you use "#{Asset Sequence Num}" as the q search parameter, it should
    only return one result. If you use a more general search term. like searching
    for the name, you may get multiple.

    :param q: The search term to search assets by.
    :type q: str
    :return: List of TokenInput objects matching the search term.
    :rtype: list[TokenInput]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/assets/items_for_token_input.json"
    response = http_get(
        url=url, params={"include_id": "true", "q": q}, title="Asset Token Input Return"
    )
    data = response.json()

    return [TokenInput(**x) for x in data]


@Decorators.check_env_vars
def asset_history_return(asset_id: int) -> list[AssetHistoryItem]:
    """
    Returns checkout history for a particular asset.

    :param asset_id: The ID of the asset to get history for.
    :type asset_id: int
    :return: List of AssetHistoryItem objects representing the asset's history.
    :rtype: list[AssetHistoryItem]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/history"

    all_history = []

    while True:
        response = http_get(url=url, title="Asset History Return")
        data = response.json()

        if "history" not in data:
            logger.error(f"Error, could not get history: {response.content}")
            raise Exception(f"Error, could not get history: {response.content}")

        all_history.extend(data["history"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [AssetHistoryItem(**x) for x in all_history]


@Decorators.check_env_vars
def asset_update(asset_id: int, update_data: dict) -> Asset | None:
    """
    Updates a fixed asset.

    :param asset_id: The ID of the asset to update.
    :type asset_id: int
    :param update_data: Dictionary of fields to update on the asset.
    :type update_data: dict
    :return: The updated asset object if successful, else None.
    :rtype: Asset | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}"
    response = http_patch(url=url, payload=update_data, title="Asset Update")

    if response.status_code == 200 and "asset" in response.json():
        return Asset(**response.json()["asset"])
    else:
        return None


# TODO Bulk update


@Decorators.check_env_vars
def asset_checkin(
    asset_id: int,
    location_id: int,
    comments: str | None = None,
    checkin_date: date | None = None,
    custom_fields: list[dict] | None = None,
) -> ResponseMessages | None:
    """
    Check in an asset to a particular location.

    :param asset_id: The ID of the asset to check in.
    :type asset_id: int
    :param location_id: The ID of the location to check the asset in to.
    :type location_id: int
    :param comments: Comments to add to the check-in.
    :type comments: str, optional
    :param checkin_date: The date to record the check-in as. Defaults to today if not provided.
    :type checkin_date: date, optional
    :param custom_fields: List of custom fields to set on check-in. Each item in
        the list should be a dictionary with 'id' and 'value' keys.
    :type custom_fields: list[dict], optional
    :return: ResponseMessages object if there are any messages, else None.
    :rtype: ResponseMessages | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/checkin"
    response = http_put(
        url=url,
        payload={
            "asset": {
                "comments": comments,
                "location_id": location_id,
                "checkin_date": checkin_date,
                "custom_fields": custom_fields,
            }
        },
        title="Asset Checkin",
    )

    if "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None


@Decorators.check_env_vars
def asset_checkout(
    asset_id: int,
    user_id: int,
    location_id: int,
    request_verification: bool,
    comments: str | None = None,
    checkout_forever: bool | None = None,
    till: datetime | None = None,
    project_id: int | None = None,
    ignore_conflicting_reservations: bool | None = None,
    fulfill_user_conflicting_reservations: bool | None = None,
    custom_fields: list[dict] | None = None,
) -> ResponseMessages | None:
    """
    Check out an asset to a member

    Note: If user is inactive, checkout will return a 200 status code but the
    asset will NOT actually be checked out. Response will contain a message.

    :param asset_id: The ID of the asset to check out.
    :type asset_id: int
    :param user_id: The ID of the user to check the asset out to.
    :type user_id: int
    :param location_id: The ID of the location to check the asset out to.
    :type location_id: int
    :param request_verification: Whether to request verification of the asset after checkout.
    :type request_verification: bool
    :param comments: Comments to add to the checkout.
    :type comments: str, optional
    :param checkout_forever: Whether to check the asset out indefinitely.
        If False, the 'till' parameter must be provided.
    :type checkout_forever: bool, optional
    :param till: The date/time to check the asset back in. Required if checkout_forever is False.
    :type till: datetime, optional
    :param project_id: The ID of the project to associate the checkout with.
    :type project_id: int, optional
    :param ignore_conflicting_reservations: Whether to ignore any conflicting reservations
        for the asset when checking out.
    :type ignore_conflicting_reservations: bool, optional
    :param fulfill_user_conflicting_reservations: Whether to fulfill any conflicting reservations
        for the user when checking out.
    :type fulfill_user_conflicting_reservations: bool, optional
    :param custom_fields: List of custom fields to set on checkout. Each item in
        the list should be a dictionary with 'id' and 'value' keys.
    :type custom_fields: list[dict], optional
    :return: ResponseMessages object if there are any messages, else None.
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/checkout"
    payload = {
        "asset": {
            "user_id": user_id,
            "comments": comments,
            "location_id": location_id,
            "request_verification": request_verification,
            "checkout_forever": checkout_forever,
            "till": till,
            "ignore_conflicting_reservations": ignore_conflicting_reservations,
            "fulfill_user_conflicting_reservations": fulfill_user_conflicting_reservations,
            "custom_fields": custom_fields,
        }
    }
    response = http_put(url=url, payload=payload, title="Asset Checkout")

    if "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None


@Decorators.check_env_vars
def asset_retire(
    asset_id: int,
    retired_on: datetime,
    retire_reason_id: int,
    salvage_value: float | None = None,
    retire_comments: str | None = None,
    location_id: int | None = None,
) -> Asset | None:
    """
    Retires an asset. Asset needs to be in an available state to retire. Essentially,
    making it no longer active/available for checking out.

    :param asset_id: The ID of the asset to retire.
    :type asset_id: int
    :param retired_on: The date to record the asset as retired.
    :type retired_on: datetime
    :param retire_reason_id: The ID of the reason for retiring the asset.
    :type retire_reason_id: int
    :param salvage_value: The salvage value of the asset.
    :type salvage_value: float, optional
    :param retire_comments: Comments to add to the retirement.
    :type retire_comments: str, optional
    :param location_id: The ID of the location to move the asset to upon retirement.
    :type location_id: int, optional
    :return: The updated asset object with retired status if successful, else None.
    :rtype: Asset | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/retire"
    payload = {
        "asset": {
            "retired_on": retired_on,
            "retire_reason_id": retire_reason_id,
            "salvage_value": salvage_value,
            "retire_comments": retire_comments,
            "location_id": location_id,
        }
    }
    response = http_put(url=url, payload=payload, title="Asset Retire")

    if "asset" in response.json():
        return Asset(**response.json()["asset"])
    else:
        return None


@Decorators.check_env_vars
def asset_activate(asset_id: int, location_id: int | None = None) -> Asset | None:
    """
    Reactivates a retired asset.

    :param asset_id: The ID of the asset to reactivate.
    :type asset_id: int
    :param location_id: The ID of the location to move the asset to upon reactivation.
        If not provided, asset will remain in its current location.
    :type location_id: int, optional
    :return: The updated asset object with active status if successful, else None.
    :rtype: Asset | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/activate"
    payload = {"asset": {"location_id": location_id}}
    response = http_put(url=url, payload=payload, title="Asset Activate")

    if "asset" in response.json():
        return Asset(**response.json()["asset"])
    else:
        return None


@Decorators.check_env_vars
def asset_verification_request(asset_id: int, note: str) -> dict:
    """
    Creates a verification request for a particular asset. Prompts the user the
    asset is currently checked out to to verify that they do still have ownership
    of it.

    :param asset_id: The ID of the asset to create a verification request for.
    :type asset_id: int
    :param note: A note to include with the verification request.
    :type note: str
    :return: The response from the API as a dictionary.
    :rtype: dict
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}/audits.json"
    # Not entirely sure on correct request body. Endpoint not yet documented in EZO v2
    # API, so just copying what I'm seeing in the browser's network tools when doing a verification request
    payload = {
        "asset_id": asset_id,
        "custom_substate_id": "",
        "audit": {
            "custom_required": "1",
            "custom_note": note,
        },
    }
    response = http_post(url=url, payload=payload, title="Asset Verification Request")

    return response.json()


@Decorators.check_env_vars
def asset_delete(asset_id: int) -> ResponseMessages | None:
    """
    Delete a particular asset.

    :param asset_id: The ID of the asset to delete.
    :type asset_id: int
    :return: ResponseMessages object if there are any messages, else None.
    :rtype: ResponseMessages | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/assets/{asset_id}"
    response = http_delete(url=url, title="Asset Delete")

    if "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None



import logging
import os
import time

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get
from ezoff.data_model import Bundle

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def bundle_create(
    name: str,
    description: str,
    identification_number: str,
    location_id: int,
    enable_items_restricted_by_location: bool,
    bundle_line_items: list[dict],
    allow_add_bundle_without_specifying_items: bool,
) -> Bundle | None:
    """
    Creates a new bundle of items.

    :param name: The name of the bundle.
    :type name: str
    :param description: A description of the bundle.
    :type description: str
    :param identification_number: A unique identification number for the bundle.
    :type identification_number: str
    :param location_id: The ID of the location the bundle is associated with.
    :type location_id: int
    :param enable_items_restricted_by_location: Whether to enable items restricted by location.
    :type enable_items_restricted_by_location: bool
    :param bundle_line_items: A list of dictionaries representing the items in the bundle.
    :type bundle_line_items: list[dict]
    :param allow_add_bundle_without_specifying_items: Whether to allow adding the bundle without specifying items.
    :type allow_add_bundle_without_specifying_items: bool
    :return: The created Bundle object if successful, else None.
    :rtype: Bundle | None
    """
    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/bundles"
    response = http_post(url=url, payload={"bundle": params}, title="Bundle Create")

    if response.status_code == 200 and "bundle" in response.json():
        return Bundle(**response.json()["bundle"])
    else:
        return None


@Decorators.check_env_vars
def bundle_return(bundle_id: int) -> Bundle | None:
    """
    Returns a particular bundle.

    :param bundle_id: The ID of the bundle to retrieve.
    :type bundle_id: int
    :return: The Bundle object if found, else None.
    :rtype: Bundle | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/bundles/{bundle_id}"
    response = http_get(url=url)

    if response.status_code == 200 and "bundle" in response.json():
        return Bundle(**response.json()["bundle"])
    else:
        return None


@Decorators.check_env_vars
def bundles_return(filter: dict | None = None) -> list[Bundle]:
    """
    Returns all bundles.

    :param filter: A dictionary of bundle fields and the values to filter by.
    :type filter: dict, optional
    :return: A list of Bundle objects.
    :rtype: list[Bundle]
    """
    if filter:
        for field in filter:
            if field not in Bundle.model_fields:
                raise ValueError(f"'{field}' is not a valid field for a bundle.")
        filter = {"filters": filter}
    else:
        filter = None

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/bundles"

    all_bundles = []

    while True:
        response = http_get(url=url, payload=filter, title="Bundles Return")
        data = response.json()

        if "bundles" not in data:
            logger.error(f"Error, could not get bundles: {response.content}")
            raise Exception(f"Error, could not get bundles: {response.content}")

        all_bundles.extend(data["bundles"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Bundle(**x) for x in all_bundles]



"""
Helper functions for ezoff
"""

import logging
import os

import requests
from tenacity import (
    after_log,
    before_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_exponential,
)

logger = logging.getLogger(__name__)


def should_retry_http_or_network_error(exception: BaseException) -> bool:
    """
    Determines if an exception warrants a retry.
    Retries on ConnectionError, Timeout, or specific HTTP 5XX errors.
    """
    if isinstance(
        exception, (requests.exceptions.ConnectionError, requests.exceptions.Timeout)
    ):
        return True
    if isinstance(exception, requests.exceptions.HTTPError):
        # Retry on 5XX server errors
        return 500 <= exception.response.status_code < 600
    return False


_basic_retry = retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    retry=retry_if_exception(should_retry_http_or_network_error),
    before=before_log(logger, logging.DEBUG),
    after=after_log(logger, logging.DEBUG),
)


@_basic_retry
def _http_request(
    call_method, url: str, title: str, headers: dict = {}, params: dict = {}, payload: dict = {}, timeout: int=60
) -> requests.Response:
    """Generic HTTP request wrapper. Used for making various types of HTTP requests to API endpoints.

    Args:
        call_method (_type_): HTTP Request method to use for call.
        url (str): Endpoint URL
        title (str): Descriptive title of call. Used in error messages.
        headers (dict, optional): HTTP header values of request.
        params (dict, optional): HTTP parameter values of request.
        payload (dict, optional): HTTP body payload of request.
        timeout (int, optional): HTTP timeout value.

    Returns:
        requests.Response: HTTP response object returned by request.
    """    
    
    def _log_request(
    ) -> None:
        """Prints request details to the error log.
        Called if an error occurrs as a result of the HTTP request.
        """

        # Redact bearer token before logging headers.
        if "Authorization" in headers:
            headers["Authorization"] = "REDACTED"

        logger.error('*' * 50)
        logger.error(msg)
        logger.error(f'HTTP Method: {call_method.__name__}')
        logger.error(f"URL: {url}")
        logger.error(f"Headers: {headers}")

        if payload is not None:
            logger.error(f"Payload: {payload}")

        if params is not None:
            logger.error(f"Params: {params}")

        logger.error('*' * 50)

    try:
        response = call_method(
            url,
            headers=headers,
            params=params,
            json=payload,
            timeout=timeout,
        )
        response.raise_for_status()

    except requests.exceptions.HTTPError as e:
        msg = f"HTTP error calling {title} API endpoint: {e.response.status_code} - {e.response.content}"
        _log_request()
        raise

    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
        msg = f"Connection error calling {title} API endpoint: {e}"
        _log_request()
        raise

    except requests.exceptions.RequestException as e:
        msg = f"Request error calling {title} API endpoint: {e}"
        _log_request()
        raise

    return response


def http_delete(
    url: str,
    title: str,
    timeout: int = 60,
    headers: dict = None,
    payload: dict = None,
    params: dict = None,
) -> requests.Response:

    if headers is None:
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['EZO_TOKEN']}",
        }

    return _http_request(
        call_method=requests.delete,
        url=url,
        headers=headers,
        params=params,
        payload=payload,
        title=title,
        timeout=timeout,
    )


def http_get(
    url: str,
    title: str,
    timeout: int = 60,
    headers: dict = None,
    payload: dict = None,
    params: dict = None,
) -> requests.Response:

    if headers is None:
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['EZO_TOKEN']}",
        }

    return _http_request(
        call_method=requests.get,
        url=url,
        headers=headers,
        params=params,
        payload=payload,
        title=title,
        timeout=timeout,
    )


def http_patch(
    url: str, title: str, timeout: int = 60, headers: dict = None, payload: dict = None
) -> requests.Response:

    if headers is None:
        headers = {
            # "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['EZO_TOKEN']}",
        }

    return _http_request(
        call_method=requests.patch,
        url=url,
        headers=headers,
        payload=payload,
        title=title,
        timeout=timeout,
    )


def http_post(
    url: str, payload: dict, title: str, timeout: int = 60, headers: dict = None
) -> requests.Response:

    if headers is None:
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['EZO_TOKEN']}",
        }

    return _http_request(
        call_method=requests.post,
        url=url,
        headers=headers,
        payload=payload,
        title=title,
        timeout=timeout,
    )


def http_put(
    url: str, payload: dict, title: str, timeout: int = 60, headers: dict = None
) -> requests.Response:

    if headers is None:
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {os.environ['EZO_TOKEN']}",
        }

    return _http_request(
        call_method=requests.put,
        url=url,
        headers=headers,
        payload=payload,
        title=title,
        timeout=timeout,
    )



"""
Module contains any enums used throughout the package.
"""

from enum import Enum


class AssetClass(Enum):
    BILL_CHANGER = "Bill Changer"
    BULK_CO2 = "Bulk CO2"
    COFFEE = "Coffee"
    COOLER = "Cooler"
    FOUNTAIN = "Fountain"
    MICRO_MARKET = "Micro Market"
    REMOVE = "Remove"
    SLUSH = "Slush"
    TEA = "Tea"
    TEST = "Test"
    VENDOR = "Vendor"
    WATER = "Water"


class CustomFieldID(Enum):
    DEPOT = 739
    EST_SVC_MINUTES = 728
    RENT_FLAG = 70779
    TAX_JURISDICTION = 738
    NAT_ACCOUNT = 739
    CANTEEONE_CODE = 740
    PARENT_CUST_CODE = 771
    EXCLUDE_RENT_FEES = 823
    ASSET_SERIAL_NO = 66133
    TELEMETRY_SERIAL_NO = 66130
    PMA_NO = 66211
    ACQUIRED_FROM = 66131
    MANUFACTURED_DATE = 20622
    MFR_PART_NO = 66212
    ASSET_CLASS = 71024
    LOCATION_CLASS = 845
    BC_VENDOR_ID = 1229
    BC_VENDOR_NAME = 1228
    BC_VENDOR_REMIT_ADDR_CODE = 1230
    BC_DEPT_ID = 1231
    PEPSICO_PARENT_OUTLET_ID = 1308


class EzoErrorMessage(str, Enum):
    COMPLETED_STATE = "you cant complete a Service Call when it is in completed state"


class LocationClass(Enum):
    NONE = "None"
    MARKET = "Micro Market"


class RentLoan(Enum):
    RENT = "Rent"
    LOAN = "Loan"


class ResourceType(Enum):
    """Ez Office component (resource) type."""

    ASSET = "Asset"



"""
Module contains any pydantic models used throughout the package.
"""

from datetime import date, datetime
from typing import Any, Literal, Optional

from ezoff.enums import AssetClass, CustomFieldID, LocationClass, ResourceType
from pydantic import BaseModel, Field


class ResponseMessages(BaseModel):
    success: list[str] | None = None
    errors: list[str] | None = None


class Asset(BaseModel):
    """
    Model representing a fixed asset. Some individual item that could be identified with
    a serial number or some unique identifier.
    """

    active_sub_checkout: Optional[Any] = Field(default=None)
    arbitration: int
    audit_pending: bool
    bulk_import_id: Optional[int] = Field(default=None)
    checkin_due_on: Optional[datetime] = Field(default=None)
    checkout_on: Optional[datetime] = Field(default=None)
    comments_count: int
    cost_price: float
    created_at: Optional[datetime] = Field(default=None)
    custom_fields: Optional[list] = Field(default=[])
    custom_substate_id: Optional[int] = Field(default=None)
    depreciation_calculation_required: bool
    description: Optional[str] = Field(default="")
    display_image: str
    documents_count: int
    group_id: int
    id: int
    identifier: str
    item_audit_id: Optional[int] = Field(default=None)
    assigned_to_id: int | None = Field(default=None)
    last_assigned_to_id: int | None = Field(default=None)
    last_checked_in_at: Optional[datetime] = Field(default=None)
    last_checked_out_at: Optional[datetime] = Field(default=None)
    last_history_id: Optional[int] = Field(default=None)
    latest_contract_id: Optional[int] = Field(default=None)
    location_id: Optional[int] = Field(default=None)
    manufacturer: Optional[str] = Field(default="")
    name: str
    package_id: Optional[int] = Field(default=None)
    pending_verification: bool
    primary_user: Optional[int] = Field(default=None)
    product_model_number: Optional[str] = Field(default="")
    purchase_order_id: Optional[int] = Field(default=None)
    purchased_on: Optional[date] = Field(default=None)
    retire_comments: Optional[str] = Field(default="")
    retire_reason_id: Optional[int] = Field(default=None)
    retired_by_id: Optional[int] = Field(default=None)
    retired_on: Optional[datetime] = Field(default=None)
    salvage_value: str
    services_count: Optional[int] = Field(default=None)
    state: str
    sub_checked_out_to_id: Optional[int] = Field(default=None)
    sub_group_id: Optional[int] = Field(default=None)
    sunshine_id: Optional[int] = Field(default=None)
    synced_with_jira_at: Optional[date] = Field(default=None)
    updated_at: Optional[datetime] = Field(default=None)
    vendor_id: Optional[int] = Field(default=None)

    # Custom fields, parsed from the custom_fields attribute.
    rent: Optional[bool] = Field(default=None)
    serial_number: Optional[str] = Field(default=None)
    telemetry_serial_number: Optional[str] = Field(default=None)
    pma_number: Optional[str] = Field(default=None)
    acquired_from: Optional[str] = Field(default=None)
    manufactured_date: Optional[str] = Field(default=None)
    mfr_part_number: Optional[str] = Field(default=None)
    asset_class: Optional[AssetClass] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        """
        Parse custom fields.
        """

        if self.custom_fields:
            # Create lookup dictionary
            field_values = {
                field.get("id"): field.get("value")
                for field in self.custom_fields
                if "id" in field
            }

            # Process Rent Flag
            if CustomFieldID.RENT_FLAG.value in field_values:
                value = field_values[CustomFieldID.RENT_FLAG.value]
                if value is not None and isinstance(value, list):
                    if len(value) > 0:
                        self.rent = True
                    else:
                        self.rent = False

            # Process Serial Number
            if CustomFieldID.ASSET_SERIAL_NO.value in field_values:
                value = field_values[CustomFieldID.ASSET_SERIAL_NO.value]
                if value and isinstance(value, str):
                    self.serial_number = value

            # Process Telemetry Serial Number
            if CustomFieldID.TELEMETRY_SERIAL_NO.value in field_values:
                value = field_values[CustomFieldID.TELEMETRY_SERIAL_NO.value]
                if value and isinstance(value, str):
                    self.telemetry_serial_number = value

            # Process PMA Number
            if CustomFieldID.PMA_NO.value in field_values:
                value = field_values[CustomFieldID.PMA_NO.value]
                if value and isinstance(value, str):
                    self.pma_number = value

            # Process Acquired From
            if CustomFieldID.ACQUIRED_FROM.value in field_values:
                value = field_values[CustomFieldID.ACQUIRED_FROM.value]
                if value and isinstance(value, str):
                    self.acquired_from = value

            # Process Manufactured Date
            if CustomFieldID.MANUFACTURED_DATE.value in field_values:
                value = field_values[CustomFieldID.MANUFACTURED_DATE.value]
                if value and isinstance(value, str):
                    self.manufactured_date = value

            # Process MFR Part Number
            if CustomFieldID.MFR_PART_NO.value in field_values:
                value = field_values[CustomFieldID.MFR_PART_NO.value]
                if value and isinstance(value, str):
                    self.mfr_part_number = value

            # Process Asset Class
            if CustomFieldID.ASSET_CLASS.value in field_values:
                value = field_values[CustomFieldID.ASSET_CLASS.value]
                if value is not None and isinstance(value, list):
                    if len(value) > 0:
                        try:
                            self.asset_class = AssetClass(value[0])
                        except ValueError as e:
                            raise ValueError(
                                f"Invalid asset class in asset {self.id}: {value[0]}"
                            )


class Inventory(BaseModel):
    arbitration: int | None = None
    asset_detail_attributes: dict | None = None
    audit_pending: bool
    available_quantity: int | None = None
    bulk_import_id: int | None = None
    comments_count: int
    cost_price: float
    created_at: datetime
    custom_fields: list[dict] | None = None
    default_excess_location_threshold: int
    default_low_location_threshold: int | None = None
    depreciation_calculation_required: bool
    description: str | None = None
    display_image: str | None = None
    documents_count: int
    group_id: int
    id: int
    identifier: str
    initial_stock_quantity: int
    inventory_threshold: int
    line_items_attributes: list | None = None
    location_id: int | None = None
    location_thresholds_attributes: list | None = None
    name: str
    net_quantity: int | None = None
    package_id: int | None = None
    pending_verification: bool
    product_model_number: str | None = None
    purchase_order_id: int | None = None
    purchased_on: datetime | None = None
    retire_comments: str | None = None
    retire_reason_id: int | None = None
    retired_on: datetime | None = None
    salvage_value: float
    state: str | None = None
    sub_group_id: int | None = None
    updated_at: datetime
    vendor_id: int | None = None


class ChecklistLineItem(BaseModel):
    title: str
    type: str


class Checklist(BaseModel):
    id: int
    name: str
    created_by_id: int
    line_items: list[ChecklistLineItem]


class Component(BaseModel):
    resource_id: int
    resource_type: ResourceType

    class Config:
        use_enum_values = True


class Location(BaseModel):
    # Class attribute to control custom fields clearing behavior
    _clear_custom_fields: bool = True

    apply_default_return_date_to_child_locations: Optional[bool] = Field(default=None)
    checkout_indefinitely: Optional[bool] = Field(default=None)
    city: Optional[str] = Field(default=None)
    comments_count: int
    country: Optional[str] = Field(default="")
    created_at: Optional[datetime] = Field(default=None)
    custom_fields: Optional[list] = Field(default=[])
    default_return_duration: Optional[int] = Field(default=None)
    default_return_duration_unit: Optional[str] = Field(default=None)
    default_return_time: Optional[datetime] = Field(default=None)
    description: Optional[str] = Field(default=None)
    documents_count: int
    hidden_on_webstore: bool
    id: int
    identification_number: Optional[str] = Field(default=None)
    latitude: Optional[float] = Field(default=None)
    longitude: Optional[float] = Field(default=None)
    manual_coordinates_provided: Optional[bool] = Field(default=None)
    name: str
    parent_id: Optional[int] = Field(default=None)
    secure_code: str
    state: Optional[str] = Field(default=None)
    status: str
    street1: Optional[str] = Field(default=None)
    street2: Optional[str] = Field(default=None)
    updated_at: Optional[datetime] = Field(default=None)
    visible_on_webstore: bool
    zip_code: Optional[str] = Field(default=None)

    # Custom fields
    parent_cust_code: Optional[str] = Field(default=None)
    exclude_rent_fees: Optional[bool] = Field(default=None)
    location_class: Optional[LocationClass] = Field(default=LocationClass.NONE)
    tax_jurisdiction: Optional[str] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        """
        Parse custom fields into specific model attributes.
        """
        if self.custom_fields:
            # Create lookup dictionary
            field_values = {
                field.get("id"): field.get("value")
                for field in self.custom_fields
                if "id" in field
            }

            # Process 'Exclude Rent Fees'
            if CustomFieldID.EXCLUDE_RENT_FEES.value in field_values:
                value = field_values[CustomFieldID.EXCLUDE_RENT_FEES.value]
                if value and isinstance(value, str):
                    self.exclude_rent_fees = value.lower() == "yes"

            # Process 'Parent Customer Code'
            if CustomFieldID.PARENT_CUST_CODE.value in field_values:
                value = field_values[CustomFieldID.PARENT_CUST_CODE.value]
                if value and isinstance(value, str):
                    self.parent_cust_code = value

            # Process 'Location Class'
            if CustomFieldID.LOCATION_CLASS.value in field_values:
                value = field_values[CustomFieldID.LOCATION_CLASS.value]
                self.location_class = LocationClass(value or LocationClass.NONE)

            # Process 'Tax Jurisdiction'
            if CustomFieldID.TAX_JURISDICTION.value in field_values:
                value = field_values[CustomFieldID.TAX_JURISDICTION.value]
                if value and isinstance(value, str):
                    self.tax_jurisdiction = value

            # Clear out custom field list, to save space (if enabled).
            if self._clear_custom_fields:
                self.custom_fields = None


class Member(BaseModel):
    """
    Model representing a member (user).
    """

    account_name: Optional[str] = Field(default=None)
    address_name: Optional[str] = Field(default=None)
    alert_type: Optional[str] = Field(default=None)
    auto_sync_with_ldap: Optional[bool] = Field(default=None)
    billing_address_id: Optional[int] = Field(default=None)
    category_id: Optional[int] = Field(default=None)
    collect_tax: Optional[str] = Field(default=None)
    comments_count: Optional[int] = Field(default=None)
    company_default_payment_terms: Optional[bool] = Field(default=None)
    contact_owner: Optional[str] = Field(default=None)
    contact_type: Optional[str] = Field(default=None)
    country: Optional[str] = Field(default=None)
    created_at: datetime
    created_by_id: Optional[int] = Field(default=None)
    creation_source: Optional[str] = Field(default=None)
    credit_memo_amount: Optional[float] = Field(default=None)
    custom_fields: Optional[list[dict]] = Field(default=[])
    deactivated_at: Optional[datetime] = Field(default=None)
    default_address_id: Optional[int] = Field(default=None)
    default_triage_setting_id: Optional[int] = Field(default=None)
    department: Optional[str] = Field(default=None)
    description: Optional[str] = Field(default=None)
    documents_count: Optional[int] = Field(default=None)
    email: str
    employee_id: Optional[str] = Field(default=None)
    employee_identification_number: Optional[str] = Field(default=None)
    fax: Optional[str] = Field(default=None)
    first_name: str | None = None
    full_name: str | None = None
    hourly_rate: Optional[float] = Field(default=None)
    id: int
    inactive_by_id: Optional[int] = Field(default=None)
    jira_account_id: Optional[str] = Field(default=None)
    last_name: str | None = None
    last_sync_date: Optional[datetime] = Field(default=None)
    last_sync_source: Optional[str] = Field(default=None)
    manager_id: Optional[int] = Field(default=None)
    offboarding_date: Optional[date] = Field(default=None)
    otp_required_for_login: Optional[bool] = Field(default=None)
    password_changed_at: Optional[datetime] = Field(default=None)
    payment_term_id: Optional[int] = Field(default=None)
    phone_number: Optional[str] = Field(default=None)
    role_id: int
    salesforce_id: Optional[int] = Field(default=None)
    secure_code: Optional[str] = Field(default=None)
    services_count: Optional[int] = Field(default=None)
    settings_access: Optional[bool] = Field(default=None)
    show_announcement: Optional[bool] = Field(default=None)
    show_app_updates: Optional[bool] = Field(default=None)
    status: int
    stock_asset_current_checkout_view: Optional[bool] = Field(default=None)
    subscribed_to_emails: Optional[bool] = Field(default=None)
    # Note: team is weird. Used to be just an int. Now, appears to be a multi-value
    # situation where you get a list of values back. However, API is inconsistent.
    # Sometimes you'll get back team_id, sometimes team_ids.
    team_id: int | list[int] | None = None
    team_ids: int | list[int] | None = None
    time_zone: Optional[str] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        """
        Normalize team_id / team_ids inconsistencies from API responses.
        Ensures team_ids is always a list[int], even if API returns a single int or null.
        """
        # Prefer team_ids if present
        if self.team_ids is not None:
            if isinstance(self.team_ids, int):
                self.team_ids = [self.team_ids]
            elif isinstance(self.team_ids, list):
                self.team_ids = [int(x) for x in self.team_ids if x is not None]
        elif self.team_id is not None:
            # Handle team_id if team_ids not provided
            if isinstance(self.team_id, int):
                self.team_ids = [self.team_id]
            elif isinstance(self.team_id, list):
                self.team_ids = [int(x) for x in self.team_id if x is not None]
            else:
                self.team_ids = []
        else:
            # Neither provided
            self.team_ids = []

        # Ensure consistency: keep team_id as first element if exists
        if self.team_ids:
            self.team_id = self.team_ids[0]
        else:
            self.team_id = None

    unseen_app_updates_count: Optional[int] = Field(default=None)
    unsubscribed_by_id: Optional[int] = Field(default=None)
    updated_at: Optional[datetime] = Field(default=None)
    user_listing_id: Optional[int] = Field(default=None)
    work_location: int | None = Field(default=None)
    zendesk_account_id: Optional[int] = Field(default=None)


class MemberCreate(BaseModel):
    """
    A model representing the data required to create a new member.
    """

    first_name: str | None = None
    last_name: str
    role_id: int
    email: str
    employee_identification_number: str | None = None
    department: str | None = None
    description: str | None = None
    team_ids: list[int] | None = None
    user_listing_id: int | None = None
    login_enabled: bool | None = None
    subscribed_to_emails: bool | None = None
    skip_confirmation_email: bool | None = None
    address_name: str | None = None
    address: str | None = None
    address_line_2: str | None = None
    city: str | None = None
    state: str | None = None
    zip_code: str | None = None
    country: str | None = None
    fax: str | None = None
    phone_number: str | None = None
    image_url: str | None = None
    work_location: int | None = None
    custom_fields: list[dict] | None = None


class CustomRole(BaseModel):
    id: int
    name: str
    description: str
    created_by_id: int
    base_role_id: int
    visibility: str
    users_visibility: str
    system_generated: bool
    group_ids: list[int] | None = None
    team_ids: list[int] | None = None
    location_ids: list[int] | None = None


class UserListing(BaseModel):
    id: int
    name: str
    default: bool
    include_sub_locations: bool
    created_at: datetime
    updated_at: datetime


class WorkOrder(BaseModel):
    class AssociatedAsset(BaseModel):
        name: str
        id: int

    approver_id: Optional[int] = Field(default=None)
    assigned_to_id: Optional[int] = Field(default=None)
    assigned_to_type: str
    asset_id: int | None = None
    associated_assets: list[AssociatedAsset] | None = None
    associated_checklists: list
    base_cost: float
    completed_on: Optional[str] = Field(default=None)
    create_one_task_for_all_items: bool
    create_recurring_service_zendesk_tickets: bool
    created_at: Optional[datetime] = Field(default=None)
    created_by_id: Optional[int] = Field(default=None)
    creation_source: Optional[str] = Field(default=None)
    custom_fields: Optional[list[dict]]
    description: Optional[str] = Field(default=None)
    display_next_service_immediately: bool
    due_date: Optional[datetime] = Field(default=None)
    expected_start_date: Optional[datetime] = Field(default=None)
    id: int
    inventory_cost: float
    inventory_cost_method: Optional[str] = Field(default=None)
    is_item_component: bool
    is_triage: bool
    location_id: Optional[int] = Field(default=None)
    mark_items_unavailable: bool
    preventive_maintenance: bool
    priority: str
    project_id: Optional[int] = Field(default=None)
    recurrence_based_on_completion_date: bool
    recurrence_task_id: Optional[int | None]
    repeat_every_basis: Optional[int] = Field(default=None)
    repeat_every_value: int
    repetition_end_date: Optional[str] = Field(default=None)
    repetition_starting: Optional[str] = Field(default=None)
    requested_by_id: Optional[int] = Field(default=None)
    require_approval_from_reviewer: bool
    reviewer_id: Optional[int] = Field(default=None)
    secondary_assignee_ids: Optional[list[int]] = Field(default=[])
    service_for_sub_groups_only: bool
    service_type_id: Optional[int] = Field(default=None)
    shipping_address_id: Optional[int] = Field(default=None)
    start_work_on_all_assets: bool
    started_on: Optional[str] = Field(default=None)
    state: str
    supervisor_id: Optional[int] = Field(default=None)
    task_type: str
    task_type_id: Optional[int] = Field(default=None)
    template_id: Optional[int] = Field(default=None)
    time_spent: float
    # time_to_respond:
    time_to_start: int
    title: str
    total_cost: float
    track_progress: float
    updated_at: str
    warranty: Optional[bool] = Field(default=False)
    work_logs_cost: float
    work_type_name: Optional[str] = Field(default=None)
    zendesk_ticket_id: Optional[int]

    # Custom fields
    depot: Optional[str] = Field(default=None)
    depot_id: Optional[int] = Field(default=None)

    def model_post_init(self, __context: Any) -> None:
        # Parse custom fields.
        if self.custom_fields:
            # Create lookup dictionary
            field_values = {
                field.get("id"): field.get("value")
                for field in self.custom_fields
                if "id" in field
            }

            # Process Depot
            if CustomFieldID.DEPOT.value in field_values:
                value = field_values[CustomFieldID.DEPOT.value]
                if value and isinstance(value, str):
                    self.depot = value
                    self.depot_id = int(value[:2])


class WorkLog(BaseModel):
    user_id: int
    time_spent: str | None = None
    work_detail: str | None = None
    associated_to_asset_id: int | None = None
    started_on: datetime | None = None
    ended_on: datetime | None = None


class DepreciationRate(BaseModel):
    id: int
    depreciation_method_id: int
    depreciation_method_name: str
    rate: str | None = None
    useful_life: int | None = None


class Group(BaseModel):
    """
    A group in EZO is a way of categorizing assets. You can have groups, as well
    as subgroups that exist in a hierarchy beneath their parent group.
    """

    id: int
    name: str
    description: str | None = None
    enable_service_triage: bool
    triage_completion_period: int | None = None
    triage_completion_period_basis: str | None = None
    allow_staff_to_set_checkout_duration: bool
    staff_checkout_duration_months: int
    staff_checkout_duration_weeks: int
    staff_checkout_duration_days: int
    staff_checkout_duration_hours: int
    staff_checkout_duration_mins: int
    available_assets_count: int
    visible_on_webstore: bool
    hidden_on_webstore: bool
    created_at: datetime
    documents_count: int
    asset_depreciation_mode: str
    comments_count: int | None = None
    depreciation_rates: list[DepreciationRate] | None = None
    parent_id: int | None = None
    group_id: int | None = None


class Team(BaseModel):
    id: int
    name: str
    description: str
    parent_id: int | None = None
    identification_number: str
    documents_count: int
    comments_count: int


class Project(BaseModel):
    id: int
    name: str
    description: str | None = None
    created_by_id: int
    state: str
    documents_count: int
    comments_count: int
    identifier: str
    linked_modules: list[
        Literal[
            "items",
            "checkouts",
            "reservations",
            "purchase_orders",
            "work_orders",
            "carts",
            "locations",
        ]
    ]


class CustomFieldHistoryItem(BaseModel):
    id: int
    value: str | None = None
    number_value: int | float | None = None
    date_value: date | None
    options_value: Any | None
    date_time_value: datetime | None = None
    line_item_id: int | None = None
    linkable_resource_value: Any | None
    created_at: datetime
    updated_at: datetime


class StockHistoryItem(BaseModel):
    id: int
    quantity: int
    price: str
    comments: str | None = None
    created_by_id: int
    basket_id: int | None = None
    is_transfer: bool
    is_custody_transfer: bool
    retire_reason_id: int | None = None
    quantity_after_transaction: int
    order_type: str
    remaining_quantity: int | None = None
    retire_reason: str | None = None
    checked_out_to_location_id: int | None = None
    checked_in_from_location_id: int | None = None
    vendor_id: int | None = None
    checkout_line_item_id: int | None = None
    purchase_order_id: int | None = None
    basket_asset_id: int | None = None
    service_id: int | None = None
    cost_price: str | None = None
    unit_cost_price: list[dict] | None
    task_id: int | None = None
    check_out_to_asset_id: int | None = None
    paired_transfer_line_item_id: int | None = None
    project_id: int | None = None
    original_quantity: int | None = None
    signed_by_name: str | None = None
    transfer_from_location_id: int | None = None
    transition_to_substate_id: int | None = None
    action_source: str | None = None
    cost_valuation_method: str | None = None
    signature_image_id: int | None = None
    item_audit_id: int | None = None
    linked_inventory_item_id: int | None = None
    parent_id: int | None = None
    agreement_document_id: int | None = None
    agreement_accepted: bool | None = None
    user_full_name: str | None = None
    for_retiring_checked_out_stock: bool | None = None
    asset_name: str | None = None
    transition_from_substate_id: int | None = None
    checkout_on: datetime | None = None
    checkin_on: datetime | None = None
    purchased_on: datetime | None = None
    checkin_due_on: datetime | None = None
    created_at: datetime
    updated_at: datetime


class Reservation(BaseModel):
    id: int
    resource_id: int
    reservable_id: int
    reservable_type: str
    from_date: datetime
    note: str | None = None
    status: str
    to_date: datetime | None = None
    action_taken_by_id: int
    created_by_id: int
    quantity: int
    location_id: int
    price: float | None = None
    package_id: int | None = None
    basket_id: int | None = None
    baskets_asset_id: int | None = None
    signature_image_id: int | None = None
    reserved_to_location_id: int | None = None
    denied_reason: str | None = None
    recurring_reservation_id: int | None = None
    transition_from_substate_id: int | None = None
    signed_by_name: str | None = None
    project_id: int | None = None
    resource_type: str | None = None
    creation_source: str | None = None
    approved_or_denied_at: datetime | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None


class TokenInput(BaseModel):
    id: int
    name: str


class AssetHistoryItem(BaseModel):
    id: int
    assigned_to_id: int
    created_by_id: int
    is_checkout: bool | None = None
    location_id: int | None = None
    comments_count: int | None = None
    package_id: int
    basket_id: int | None = None
    is_transfer: bool | None = None
    checked_out_duration_in_seconds: int
    rent_collected: str | None = None
    signed_by_name: str | None = None
    signature_image_id: str | None = None
    action_source: str | None = None
    past_checkout: bool
    assigned_to_type: str
    basket_asset_id: int | None = None
    agreement_document_id: int | None = None
    agreement_accepted: bool | None = None
    assigned_to_name: str | None = None
    assigned_asset: str | None = None
    project_id: int | None = None
    checkin_due_on: datetime | str | None = None
    actual_checkin_on: datetime | str | None = None
    checkin_on: datetime | str | None = None
    checkout_on: datetime | None = None
    created_at: datetime
    updated_at: datetime


class Bundle(BaseModel):
    id: int
    name: str
    description: str | None = None
    identification_number: str | None = None
    location_id: int | None = None
    documents_count: int | None = None
    comments_count: int | None = None
    state: str
    enable_items_restricted_by_location: bool
    allow_add_bundle_without_specifying_items: bool
    created_at: datetime
    updated_at: datetime
    line_items: list[dict]
    custom_fields: list[dict]


class PurchaseOrder(BaseModel):
    id: int
    description: str | None = None
    title: str | None = None
    identification_number: str | None = None
    requested_by_id: int | None = None
    created_by_id: int | None = None
    approved_by_id: int | None = None
    approver_type: str | None = None
    approver_id: int | None = None
    payment_terms: str | None = None
    notes: str | None = None
    shipment_terms: str | None = None
    vendor_id: int | None = None
    state: str | None = None
    net_amount: str | None = None
    payable_amount: str | None = None
    paid_amount: str | None = None
    documents_count: int | None = None
    comments_count: int | None = None
    tax_amounts: dict | None = None
    po_type: str | None = None
    project_id: int | None = None
    contract_id: int | None = None
    receiving_notes: str | None = None
    invoice_number: str | None = None
    time_to_respond: int | None = None
    currency_id: int | None = None
    delivery_location_id: int | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    confirmed_at: datetime | None = None
    requested_on: datetime | None = None
    completed_on: datetime | None = None
    delivery_date: datetime | None = None
    custom_fields: list[dict]
    line_items: list[dict]


class Package(BaseModel):
    package_id: int
    name: str
    asset_ids: list[int]
    description: str | None = None
    arbitration: str | None = None


class RetireReason(BaseModel):
    id: int
    name: str
    active: bool
    system_generated: bool
    include_in_shrinkage: bool
    created_at: datetime
    updated_at: datetime


class Vendor(BaseModel):
    id: int
    name: str
    address: str | None = None
    contact_person_name: str | None = None
    website: str | None = None
    description: str | None = None
    email: str | None = None
    fax: str | None = None
    phone: str | None = None
    status: bool
    custom_fields: list[dict]


class StockAsset(BaseModel):
    name: str
    description: str | None = None
    cost_price: str | None = None
    identifier: str | None = None
    audit_pending: bool
    product_model_number: str | None = None
    documents_count: int
    pending_verification: bool
    comments_count: int
    state: str | None = None
    salvage_value: str | None = None
    vendor_id: int | None = None
    group_id: int
    bulk_import_id: int | None = None
    location_id: int | None = None
    depreciation_calculation_required: bool
    package_id: int | None = None
    purchase_order_id: int | None = None
    retire_comments: str | None = None
    retire_reason_id: int | None = None
    sub_group_id: int | None = None
    arbitration: int | None = None
    last_history_id: int | None = None
    retired_by_id: int | None = None
    item_audit_id: int | None = None
    sub_checked_out_to_id: int | None = None
    last_assigned_to_id: int | None = None
    manufacturer: str | None = None
    sunshine_id: int | None = None
    latest_contract_id: int | None = None
    custom_substate_id: int | None = None
    id: int
    display_image: str | None = None
    primary_user: str | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    checkin_due_on: datetime | None = None
    retired_on: datetime | None = None
    checkout_on: datetime | None = None
    purchased_on: datetime | None = None
    last_checked_out_at: datetime | None = None
    last_checked_in_at: datetime | None = None
    synced_with_jira_at: datetime | None = None
    custom_fields: list[dict]
    net_quantity: int | None = None
    available_quantity: int | None = None
    inventory_threshold: int | None = None
    location_based_threshold: int | None = None
    initial_stock_quantity: int | None = None
    default_excess_location_threshold: int | None = None
    average_cost_per_unit: str | None = None


class LinkedInventory(BaseModel):
    inventory_id: int
    quantity: int
    id: int | None = None
    location_id: int | None = None
    work_order_id: int | None = None
    asset_id: int | None = None
    cost_price: str | None = None
    original_quantity: int | None = None
    resource_id: int | None = None
    resource_type: str | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None



import logging
import os
import time

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get, http_patch
from ezoff.data_model import Vendor

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def vendor_create(
    name: str,
    address: str | None = None,
    description: str | None = None,
    email: str | None = None,
    fax: str | None = None,
    phone: str | None = None,
    website: str | None = None,
    contact_person_name: str | None = None,
    status: bool | None = None,
    custom_fields: list[dict] | None = None,
) -> Vendor | None:
    """
    Creates a new vendor.

    :param name: The name of the vendor.
    :type name: str
    :param address: The address of the vendor.
    :type address: str, optional
    :param description: A description of the vendor.
    :type description: str, optional
    :param email: The email address of the vendor.
    :type email: str, optional
    :param fax: The fax number of the vendor.
    :type fax: str, optional
    :param phone: The phone number of the vendor.
    :type phone: str, optional
    :param website: The website of the vendor.
    :type website: str, optional
    :param contact_person_name: The name of the contact person for the vendor.
    :type contact_person_name: str, optional
    :param status: The status of the vendor. True for active, False for inactive.
    :type status: bool, optional
    :param custom_fields: List of custom fields to set on the vendor. Each item in
        the list should be a dictionary with 'id' and 'value' keys.
    :type custom_fields: list[dict], optional
    :return: The created vendor object if successful, else None.
    :rtype: Vendor | None
    """
    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/vendors"
    response = http_post(url=url, payload={"vendor": params}, title="Vendor Create")

    if response.status_code == 200 and "vendor" in response.json():
        return Vendor(**response.json()["vendor"])
    else:
        return None


@Decorators.check_env_vars
def vendor_return(vendor_id: int) -> Vendor | None:
    """
    Returns a particular vendor.

    :param vendor_id: The ID of the vendor to return.
    :type vendor_id: int
    :return: The vendor object if found, else None.
    :rtype: Vendor | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/vendors/{vendor_id}"
    response = http_get(url=url, title="Vendor Return")

    if response.status_code == 200 and "vendor" in response.json():
        return Vendor(**response.json()["vendor"])
    else:
        return None


@Decorators.check_env_vars
def vendors_return() -> list[Vendor]:
    """
    Returns all vendors.

    :return: List of all vendor objects.
    :rtype: list[Vendor]
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/vendors"

    all_vendors = []

    while True:
        response = http_get(url=url, title="Vendors Return")
        data = response.json()

        if "vendors" not in data:
            logger.error(f"Error, could not get vendors: {response.content}")
            raise Exception(f"Error, could not get vendors: {response.content}")

        all_vendors.extend(data["vendors"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Vendor(**x) for x in all_vendors]


@Decorators.check_env_vars
def vendor_update(vendor_id: int, update_data: dict) -> Vendor | None:
    """
    Updates a particular vendor.

    :param vendor_id: The ID of the vendor to update.
    :type vendor_id: int
    :param update_data: A dictionary of fields to update on the vendor.
    :type update_data: dict
    :return: The updated vendor object if successful, else None.
    :rtype: Vendor | None
    """
    for field in update_data:
        if field not in Vendor.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a vendor.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/vendors/{vendor_id}"
    response = http_patch(
        url=url, payload={"vendor": update_data}, title="Vendor Update"
    )

    if response.status_code == 200 and "vendor" in response.json():
        return Vendor(**response.json()["vendor"])
    else:
        return None



"""
This module contains functions for interacting with locations in EZOfficeInventory
"""

import logging
import os
import time
from typing import Literal

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get, http_patch
from ezoff.data_model import Location

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def location_create(
    name: str,
    city: str | None = None,
    status: str | None = None,
    street1: str | None = None,
    street2: str | None = None,
    state: str | None = None,
    zip_code: str | None = None,
    description: str | None = None,
    parent_id: int | None = None,
    latitude: int | None = None,
    longitude: int | None = None,
    country: str | None = None,
    identification_number: str | None = None,
    manual_coordinates_provided: bool | None = None,
    default_return_duration: int | None = None,
    default_return_deuration_unit: str | None = None,
    default_return_time: str | None = None,
    apply_default_return_date_to_child_locations: bool | None = None,
    custom_fields: list[dict] | None = None,
) -> Location | None:
    """
    Creates a new location.

    :param name: Name of the location
    :type name: str
    :param city: City where the location is situated
    :type city: str, optional
    :param status: Status of the location (e.g., 'active', 'inactive')
    :type status: str, optional
    :param street1: First line of the street address
    :type street1: str, optional
    :param street2: Second line of the street address
    :type street2: str, optional
    :param state: State where the location is situated
    :type state: str, optional
    :param zip_code: ZIP code of the location
    :type zip_code: str, optional
    :param description: Description of the location
    :type description: str, optional
    :param parent_id: ID of the parent location, if any
    :type parent_id: int, optional
    :param latitude: Latitude coordinate of the location
    :type latitude: int, optional
    :param longitude: Longitude coordinate of the location
    :type longitude: int, optional
    :param country: Country where the location is situated
    :type country: str, optional
    :param identification_number: Identification number for the location
    :type identification_number: str, optional
    :param manual_coordinates_provided: Whether manual coordinates are provided
    :type manual_coordinates_provided: bool, optional
    :param default_return_duration: Default return duration for items at this location
    :type default_return_duration: int, optional
    :param default_return_deuration_unit: Unit for the default return duration (e.g., 'days', 'weeks')
    :type default_return_deuration_unit: str, optional
    :param default_return_time: Default return time for items at this location
    :type default_return_time: str, optional
    :param apply_default_return_date_to_child_locations: Whether to apply default return date to child
    :type apply_default_return_date_to_child_locations: bool, optional
    :param custom_fields: List of custom fields for the location
    :type custom_fields: list of dict, optional
    :return: The created location, or None if creation failed
    :rtype: Location | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = (
        f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations"
    )
    payload = {"location": params}
    response = http_post(url=url, payload=payload, title="Location Create")

    if response.status_code == 200 and "location" in response.json():
        return Location(**response.json()["location"])
    else:
        return None


@Decorators.check_env_vars
def location_return(location_id: int) -> Location | None:
    """
    Returns a particular location.

    :param location_id: The ID of the location to return
    :type location_id: int
    :return: The location with the specified ID, or None if not found
    :rtype: Location | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations/{location_id}"
    response = http_get(url=url, title="Location Return")

    if response.status_code == 200 and "location" in response.json():
        return Location(**response.json()["location"])
    else:
        return None


@Decorators.check_env_vars
def locations_return(
    state: Literal["active", "inactive"] | None = None,
    filter: dict = None,
) -> list[Location]:
    """
    Returns all locations. Optionally filter by state (active, inactive).
    Note: Unfortunately, EZO doesn't appear to have any further filtering options
    for the locations endpoint, only state. Otherwise, a more generic filter_data
    parameter would be used here.
    Optional filter parameter is for compatibility with ezo_cache class.

    :param state: Filter locations by state ('active', 'inactive')
    :type state: str, optional
    :param filter: Raw filter json for EZO API.
    :type filter: dict, optional
    :return: A list of all locations
    :rtype: list of Location
    """
    if state is not None and filter is not None:
        raise ValueError(
            "State and filter are mutually exclusive options for ezoff.locations_return()"
        )

    if state is not None:
        filter_data = {"filters": {"state": state}}
    elif filter is not None:
        filter_data = {"filters": filter}
    else:
        filter_data = None

    url = (
        f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations"
    )

    all_locations = []
    while True:
        response = http_get(url=url, payload=filter_data, title="Locations Return")
        data = response.json()

        if "locations" not in data:
            logger.error(f"Error, could not get locations: {data}")
            raise Exception(f"Error, could not get locations: {response.content}")

        all_locations.extend(data["locations"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        # Potentially running into rate limiting issues with this endpoint
        # Sleep for a second to avoid this
        time.sleep(1)

    return [Location(**x) for x in all_locations]


@Decorators.check_env_vars
def location_activate(
    location_id: int, activate_children: bool | None = None
) -> Location | None:
    """
    Activates a particular location.

    :param location_id: The ID of the location to activate
    :type location_id: int
    :param activate_children: Whether to activate all child locations as well
    :type activate_children: bool, optional
    :return: The activated location, or None if activation failed
    :rtype: Location | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations/{location_id}/activate"

    if activate_children:
        data = {"location": {"activate_all_children_locations": True}}
    else:
        data = None

    response = http_patch(url=url, payload=data, title="Location Activate")

    if response.status_code == 200 and "location" in response.json():
        return Location(**response.json()["location"])
    else:
        return None


@Decorators.check_env_vars
def location_deactivate(location_id: int) -> Location | None:
    """
    Deactivates a particular location.

    :param location_id: The ID of the location to deactivate
    :type location_id: int
    :return: The deactivated location, or None if deactivation failed
    :rtype: Location | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations/{location_id}/deactivate"
    response = http_patch(url=url, title="Location Deactivate")

    if response.status_code == 200 and "location" in response.json():
        return Location(**response.json()["location"])
    else:
        return None


@Decorators.check_env_vars
def location_update(location_id: int, update_data: dict) -> Location | None:
    """
    Updates a particular location.

    :param location_id: The ID of the location to update
    :type location_id: int
    :param update_data: A dictionary of fields to update and their new values
    :type update_data: dict
    :return: The updated location, or None if the update failed
    :rtype: Location | None
    """

    for field in update_data:
        if field not in Location.model_fields:
            raise ValueError(f"'{field}' is not a valid field for a location.")

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/locations/{location_id}"
    payload = {"location": update_data}
    response = http_patch(url=url, payload=payload, title="Location Deactivate")

    if response.status_code == 200 and "location" in response.json():
        return Location(**response.json()["location"])
    else:
        return None



"""
Covers everything related to inventory assets.
"""

import logging
import os
import time
from datetime import datetime

from ezoff._auth import Decorators
from ezoff._helpers import (
    http_post,
    http_put,
    http_get,
    http_patch,
    http_delete,
)
from ezoff.data_model import (
    CustomFieldHistoryItem,
    Inventory,
    Reservation,
    ResponseMessages,
    StockHistoryItem,
)

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def inventory_create(
    name: str,
    group_id: int,
    location_id: int,
    display_image: str | None = None,
    identifier: str | None = None,
    description: str | None = None,
    product_model_number: str | None = None,
    cost_price: float | None = None,
    vendor_id: int | None = None,
    salvage_value: float | None = None,
    sub_group_id: int | None = None,
    inventory_treshold: int | None = None,
    default_low_location_threshold: int | None = None,
    default_excess_location_threshold: int | None = None,
    initial_stock_quantity: int | None = None,
    line_item_atributes: list[dict] | None = None,
    location_thresholds_attributes: list[dict] | None = None,
    asset_detail_attributes: dict | None = None,
    custom_fields: list[dict] | None = None,
) -> Inventory | None:
    """
    Creates a new inventory item.

    :param name: Name of the inventory item
    :type name: str
    :param group_id: ID of the group the inventory item belongs to
    :type group_id: int
    :param location_id: ID of the default location where the inventory item is stored
    :type location_id: int
    :param display_image: URL of the display image for the inventory item
    :type display_image: str, optional
    :param identifier: Unique identifier for the inventory item
    :type identifier: str, optional
    :param description: Description of the inventory item
    :type description: str, optional
    :param product_model_number: Product model number of the inventory item
    :type product_model_number: str, optional
    :param cost_price: Cost price of the inventory item
    :type cost_price: float, optional
    :param vendor_id: ID of the vendor associated with the inventory item
    :type vendor_id: int, optional
    :param salvage_value: Salvage value of the inventory item
    :type salvage_value: float, optional
    :param sub_group_id: ID of the subgroup the inventory item belongs to
    :type sub_group_id: int, optional
    :param inventory_threshold: Inventory threshold for the item
    :type inventory_threshold: int, optional
    :param default_low_location_threshold: Default low location threshold for the item
    :type default_low_location_threshold: int, optional
    :param default_excess_location_threshold: Default excess location threshold for the item
    :type default_excess_location_threshold: int, optional
    :param initial_stock_quantity: Initial stock quantity of the inventory item
    :type initial_stock_quantity: int, optional
    :param line_item_attributes: List of line item attributes for the inventory item
    :type line_item_attributes: list of dict, optional
    :param location_thresholds_attributes: List of location threshold attributes for the inventory item
    :type location_thresholds_attributes: list of dict, optional
    :param asset_detail_attributes: Dictionary of asset detail attributes for the inventory item
    :type asset_detail_attributes: dict, optional
    :param custom_fields: List of custom fields for the inventory item
    :type custom_fields: list of dict, optional
    :return: The created inventory item, or None if creation failed
    :rtype: Inventory | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = (
        f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory"
    )
    response = http_post(
        url=url, payload={"inventory": params}, title="Inventory Create"
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_return(inventory_id: int) -> Inventory | None:
    """
    Get details for a particular inventory item.

    :param inventory_id: The ID of the inventory item to retrieve
    :type inventory_id: int
    :return: The inventory item with the specified ID, or None if not found
    :rtype: Inventory | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}"
    response = http_get(url=url, title="Inventory Return")

    if "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventories_return(filter: dict | None = None) -> list[Inventory]:
    """
    Returns all inventory items.

    :param filter: A dictionary of fields and their values to filter the inventory items by
    :type filter: dict, optional
    :return: A list of all inventory items matching the filter
    :rtype: list[Inventory]
    """

    if filter:
        for field in filter:
            if field not in Inventory.model_fields:
                raise ValueError(f"'{field}' is not a valid field for an inventory.")
        filter = {"filters": filter}
    else:
        filter = None

    url = (
        f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory"
    )

    all_inventories = []

    while True:
        response = http_get(url=url, payload=filter, title="Inventories Return")
        data = response.json()

        if "inventory" not in data:
            logger.error(f"Error, could not get inventories: {response.content}")
            raise Exception(f"Error, could not get inventories: {response.content}")

        all_inventories.extend(data["inventory"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Inventory(**x) for x in all_inventories]


@Decorators.check_env_vars
def inventories_search(search_term: str) -> list[Inventory]:
    """
    Searches for inventory items. Largely equivalent to the search box in the EZO UI.
    Generally recommended to use inventories_return with a filter if you have any
    sort of specific criteria to go off of. But search can be useful
    for more general queries based off user input.

    :param search_term: The term to search for in inventory items
    :type search_term: str
    :return: A list of inventory items matching the search term
    :rtype: list[Inventory]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/search"

    all_inventories = []

    while True:
        response = http_get(
            url=url, payload={"search": search_term}, title="Inventories Search"
        )
        data = response.json()

        if "inventories" not in data:
            logger.error(f"Error, could not get inventories: {response.content}")
            raise Exception(f"Error, could not get inventories: {response.content}")

        all_inventories.extend(data["inventories"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Inventory(**x) for x in all_inventories]


@Decorators.check_env_vars
def inventory_add_stock(
    inventory_id: int,
    location_id: int,
    quantity: int,
    total_price: float,
    purchased_on: datetime | None = None,
    order_by_id: int | None = None,
    vendor_id: int | None = None,
    comments: str | None = None,
    custom_fields: list[dict] | None = None,
) -> Inventory | None:
    """
    Adds stock to an inventory item at a specific location.

    :param inventory_id: The ID of the inventory item to add stock to
    :type inventory_id: int
    :param location_id: The ID of the location where the stock is being added
    :type location_id: int
    :param quantity: The quantity of stock to add
    :type quantity: int
    :param total_price: The total price of the stock being added
    :type total_price: float
    :param purchased_on: The date the stock was purchased
    :type purchased_on: datetime, optional
    :param order_by_id: The ID of the user who ordered the stock
    :type order_by_id: int, optional
    :param vendor_id: The ID of the vendor from whom the stock was purchased
    :type vendor_id: int, optional
    :param comments: Any comments regarding the stock addition
    :type comments: str, optional
    :param custom_fields: A list of custom fields to associate with the stock addition
    :type custom_fields: list of dict, optional
    :return: The updated inventory item, or None if the addition failed
    :rtype: Inventory | None
    """

    params = {k: v for k, v in locals().items() if v is not None}
    params.pop("inventory_id", None)

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/add_stock"
    response = http_post(
        url=url, payload={"inventory": params}, title="Inventory Add Stock"
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_remove_stock(
    inventory_id: int,
    location_id: int,
    to_location_id: int,
    quantity: int,
    total_price: float,
    purchased_on: datetime | None = None,
    order_by_id: int | None = None,
    vendor_id: int | None = None,
    comments: str | None = None,
    ignore_conflicting_reservations: bool | None = None,
    custom_fields: list[dict] | None = None,
) -> Inventory | None:
    """
    Removes stock of inventory item at a specific location.

    :param inventory_id: The ID of the inventory item to remove stock from
    :type inventory_id: int
    :param location_id: The ID of the location from which the stock is being removed
    :type location_id: int
    :param to_location_id: The ID of the location to which the stock is being moved (if applicable)
    :type to_location_id: int | None
    :param quantity: The quantity of stock to remove
    :type quantity: int
    :param total_price: The total price of the stock being removed
    :type total_price: float
    :param purchased_on: The date the stock was purchased
    :type purchased_on: datetime, optional
    :param order_by_id: The ID of the user who ordered the stock
    :type order_by_id: int, optional
    :param vendor_id: The ID of the vendor from whom the stock was purchased
    :type vendor_id: int, optional
    :param comments: Any comments regarding the stock removal
    :type comments: str, optional
    :param ignore_conflicting_reservations: Whether to ignore conflicting reservations when removing stock
    :type ignore_conflicting_reservations: bool, optional
    :param custom_fields: A list of custom fields to associate with the stock removal
    :type custom_fields: list of dict, optional
    :return: The updated inventory item, or None if the removal failed
    :rtype: Inventory | None
    """

    params = {k: v for k, v in locals().items() if v is not None}
    params.pop("inventory_id", None)

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/remove_stock"
    response = http_post(
        url=url, payload={"inventory": params}, title="Inventory Remove Stock"
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_update_location(inventory_id: int, location_id: int) -> Inventory | None:
    """
    Updates default location of inventory item.

    :param inventory_id: The ID of the inventory item to update
    :type inventory_id: int
    :param location_id: The ID of the new default location for the inventory item
    :type location_id: int
    :return: The updated inventory item, or None if the update failed
    :rtype: Inventory | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/update_location"
    response = http_patch(
        url=url,
        payload={"inventory": {"location_id": location_id}},
        title="Inventory Update Location",
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_transfer_stock(
    inventory_id: int,
    from_location_id: int,
    to_location_id: int,
    quantity: int,
    total_price: float,
    comments: str | None = None,
    custom_fields: list[dict] | None = None,
):
    """
    Transfers inventory item amount from one location to another.

    :param inventory_id: The ID of the inventory item to transfer stock for
    :type inventory_id: int
    :param from_location_id: The ID of the location from which the stock is being transferred
    :type from_location_id: int
    :param to_location_id: The ID of the location to which the stock is being transferred
    :type to_location_id: int
    :param quantity: The quantity of stock to transfer
    :type quantity: int
    :param total_price: The total price of the stock being transferred
    :type total_price: float
    :param comments: Any comments regarding the stock transfer
    :type comments: str, optional
    :param custom_fields: A list of custom fields to associate with the stock transfer
    :type custom_fields: list of dict, optional
    :return: The updated inventory item, or None if the transfer failed
    :rtype: Inventory | None
    """

    params = {k: v for k, v in locals().items() if v is not None}
    params.pop("inventory_id", None)

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/transfer_stock"
    response = http_post(
        url=url, payload={"inventory": params}, title="Inventory Transfer Stock"
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_retire(
    inventory_id: int,
    retire_reason_id: int,
    salvage_value: float | None = None,
    retire_comments: str | None = None,
    location_id: int | None = None,
):
    """
    Retires an inventory item.

    :param inventory_id: The ID of the inventory item to retire
    :type inventory_id: int
    :param retire_reason_id: The ID of the reason for retiring the inventory item
    :type retire_reason_id: int
    :param salvage_value: The salvage value of the inventory item
    :type salvage_value: float, optional
    :param retire_comments: Any comments regarding the retirement of the inventory item
    :type retire_comments: str, optional
    :param location_id: The ID of the location from which the inventory item is being retired
    :type location_id: int, optional
    :return: The updated inventory item, or None if the retirement failed
    :rtype: Inventory | None
    """

    params = {k: v for k, v in locals().items() if v is not None}
    params.pop("inventory_id", None)

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/retire"
    response = http_put(
        url=url, payload={"inventory": params}, title="Inventory Retire"
    )

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_activate(inventory_id: int):
    """
    Reactivates a retired inventory item.

    :param inventory_id: The ID of the inventory item to reactivate
    :type inventory_id: int
    :return: The updated inventory item, or None if the reactivation failed
    :rtype: Inventory | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/activate"
    response = http_put(url=url, title="Inventory Activate")

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_delete(inventory_id: int):
    """
    Deletes a particular inventory item.

    :param inventory_id: The ID of the inventory item to delete
    :type inventory_id: int
    :return: The deleted inventory item, or None if the deletion failed
    :rtype: Inventory | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}"
    response = http_delete(url=url, title="Inventory Delete")

    if response.status_code == 200 and "inventory" in response.json():
        return Inventory(**response.json()["inventory"])
    else:
        return None


@Decorators.check_env_vars
def inventory_quantity_by_location_return(
    inventory_id: int, location_id: int
) -> int | None:
    """
    Gets the current quantity of an inventory item in a particular location.

    :param inventory_id: The ID of the inventory item to check
    :type inventory_id: int
    :param location_id: The ID of the location to check
    :type location_id: int
    :return: The quantity of the inventory item in the specified location, or None if not found
    :rtype: int | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/get_quantity_by_location"
    response = http_get(
        url=url,
        payload={"location_id": location_id},
        title="Inventory Qty by Location Return",
    )

    if response.status_code != 200 or "quantity" not in response.json():
        return None

    return response.json()["quantity"]


# @Decorators.check_env_vars
# def inventory_line_item_locations_return(inventory_id: int):
#     """
#     TODO Doesn't seem to return useful info when testing this endpoint.
#     Just returns a listing of all locations, not anything specific to the
#     particular item.
#     """
#     pass


@Decorators.check_env_vars
def inventory_custom_field_history_return(
    inventory_id: int, custom_field_id: int
) -> list[CustomFieldHistoryItem]:
    """
    Returns custom attribute history for a particulary inventory item.

    :param inventory_id: The ID of the inventory item to retrieve custom field history for
    :type inventory_id: int
    :param custom_field_id: The ID of the custom field to retrieve history for
    :type custom_field_id: int
    :return: A list of custom field history items for the specified inventory item and custom field
    :rtype: list[CustomFieldHistoryItem]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/custom_field_history/{custom_field_id}"

    all_custom_history = []

    while True:
        response = http_get(url=url, title="Inventory Custom Field History Return")
        data = response.json()

        if "custom_field" not in data:
            logger.error(
                f"Error, could not get custom attribute history: {response.content}"
            )
            raise Exception(
                f"Error, could not get custom attribute history: {response.content}"
            )

        all_custom_history.extend(data["custom_field"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [CustomFieldHistoryItem(**x) for x in all_custom_history]


# @Decorators.check_env_vars
# def inventory_location_based_threshold_return():
#     """
#     TODO Doesn't seem to exist. Just getting a 404 when testing in Postman
#     """
#     pass


@Decorators.check_env_vars
def inventory_history_return(inventory_id: int) -> list[StockHistoryItem]:
    """
    Gets stock history of an inventory item.

    :param inventory_id: The ID of the inventory item to retrieve stock history for
    :type inventory_id: int
    :return: A list of stock history items for the specified inventory item
    :rtype: list[StockHistoryItem]
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/history"

    all_stock_history = []

    while True:
        response = http_get(
            url=url,
            title="Inventory History Return",
        )
        data = response.json()

        if "stock_history" not in data:
            logger.error(f"Error, could not get stock history: {response.content}")
            raise Exception(f"Error, could not get stock history: {response.content}")

        all_stock_history.extend(data["stock_history"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [StockHistoryItem(**x) for x in all_stock_history]


@Decorators.check_env_vars
def inventory_reservations_return(inventory_id: int) -> list[Reservation]:
    """
    Returns all reservations on an inventory item.

    :param inventory_id: The ID of the inventory item to retrieve reservations for
    :type inventory_id: int
    :return: A list of reservations for the specified inventory item
    :rtype: list[Reservation]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/{inventory_id}/reservations"

    all_reservations = []

    while True:
        response = http_get(
            url=url,
            title="Inventory Reservations Return",
        )
        data = response.json()

        if "reservations" not in data:
            logger.error(f"Error, could not get reservations: {response.content}")
            raise Exception(f"Error, could not get reservations: {response.content}")

        all_reservations.extend(data["reservations"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Reservation(**x) for x in all_reservations]


@Decorators.check_env_vars
def inventory_link_to_project(
    project_id: int, inventory_ids: list[int]
) -> ResponseMessages | None:
    """
    Links one or more inventory items to a project.

    :param project_id: ID of the project to link the inventory items to
    :type project_id: int
    :param inventory_ids: List of inventory item IDs to link to the project
    :type inventory_ids: list of int
    :return: Response messages if the linking was successful, or None if it failed
    :rtype: ResponseMessages | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/link_to_project"
    response = http_post(
        url=url,
        payload={"project_id": project_id, "ids": inventory_ids},
        title="Inventory Link to Project",
    )

    if response.status_code == 200 and "messages" in response.json():
        return ResponseMessages(**response.json()["messages"])
    else:
        return None


@Decorators.check_env_vars
def inventory_unlink_from_project(project_id: int, inventory_ids: list[int]):
    """
    Unlink one or more inventory items from a project.

    :param project_id: ID of the project to unlink the inventory items from
    :type project_id: int
    :param inventory_ids: List of inventory item IDs to unlink from the project
    :type inventory_ids: list of int
    :return: Response messages if the unlinking was successful, or None if it failed
    :rtype: ResponseMessages | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/inventory/unlink_from_project"
    response = http_post(
        url=url,
        payload={"project_id": project_id, "ids": inventory_ids},
        title="Inventory UnLink from Project",
    )

    if response.status_code == 200 and "messages" in response.json(0):
        return ResponseMessages(**response.json()["messages"])
    else:
        return None



"""
Projects in EZOffice
"""

import logging
import os
import time
from datetime import date
from typing import Literal

from ezoff._auth import Decorators
from ezoff._helpers import http_post, http_get, http_patch
from ezoff.data_model import Project

logger = logging.getLogger(__name__)


@Decorators.check_env_vars
def project_create(
    name: str,
    description: str | None = None,
    identifier: str | None = None,
    expected_start_date: date | None = None,
    expected_end_date: date | None = None,
    linked_modules: (
        Literal[
            "items",
            "checkouts",
            "reservations",
            "purchase_orders",
            "work_orders",
            "carts",
            "locations",
        ]
        | None
    ) = None,
    assigned_user_ids: list[int] | None = None,
) -> Project | None:
    """
    Creates a new project.

    :param name: Name of the project.
    :type name: str
    :param description: Description of the project.
    :type description: str, optional
    :param identifier: Identifier for the project.
    :type identifier: str, optional
    :param expected_start_date: Expected start date of the project.
    :type expected_start_date: date, optional
    :param expected_end_date: Expected end date of the project.
    :type expected_end_date: date, optional
    :param linked_modules: Modules to link to the project.
    :type linked_modules: str, optional
    :param assigned_user_ids: User IDs to assign to the project.
    :type assigned_user_ids: list of int, optional
    :return: The created project or None if creation failed.
    :rtype: Project | None
    """

    params = {k: v for k, v in locals().items() if v is not None}

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/projects"
    response = http_post(url=url, payload={"project": params}, title="Project Create")

    if response.status_code == 200 and "project" in response.json():
        return Project(**response.json()["project"])
    else:
        return None


@Decorators.check_env_vars
def project_return(project_id: int) -> Project | None:
    """
    Returns a particular project.

    :param project_id: ID of the project to return.
    :type project_id: int
    :return: The requested project or None if not found.
    :rtype: Project | None
    """
    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/projects/{project_id}"
    response = http_get(url=url, title="Project Return")

    if response.status_code == 200 and "project" in response.json():
        return Project(**response.json()["project"])
    else:
        return None


@Decorators.check_env_vars
def projects_return() -> list[Project]:
    """
    Returns all projects.

    :return: A list of all projects.
    :rtype: list[Project]
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/projects"

    all_projects = []

    while True:
        response = http_get(url=url, title="Projects Return")
        data = response.json()

        if "projects" not in data:
            logger.error(f"Error, could not get projects: {response.content}")
            raise Exception(f"Error, could not get projects: {response.content}")

        all_projects.extend(data["projects"])

        if (
            "metadata" not in data
            or "next_page" not in data["metadata"]
            or data["metadata"]["next_page"] is None
        ):
            break

        # Get the next page's url from the current page of data.
        url = data["metadata"]["next_page"]

        time.sleep(1)

    return [Project(**x) for x in all_projects]


# @Decorators.check_env_vars
# def project_linked_items_return(project_id: int):
#     """
#     Returns objects for a given module linked to a project.
#     # TODO API endpoint seems to just 500 internal server error
#     # in my testing in Postman.
#     """


@Decorators.check_env_vars
def project_mark_complete(project_id: int) -> Project | None:
    """
    Mark a project as complete.

    :param project_id: ID of the project to mark as complete.
    :type project_id: int
    :return: The updated project or None if marking complete failed.
    :rtype: Project | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/projects/{project_id}/mark_complete"
    response = http_patch(url=url, title="Project Mark Complete")

    if response.status_code == 200 and "project" in response.json():
        return Project(**response.json()["project"])
    else:
        return None


@Decorators.check_env_vars
def project_mark_in_progress(project_id: int) -> Project | None:
    """
    Mark a project as in progress.

    :param project_id: ID of the project to mark as in progress.
    :type project_id: int
    :return: The updated project or None if marking in progress failed.
    :rtype: Project | None
    """

    url = f"https://{os.environ['EZO_SUBDOMAIN']}.ezofficeinventory.com/api/v2/projects/{project_id}/mark_in_progress"
    response = http_patch(url=url, title="Project Mark In-Progress")

    if response.status_code == 200 and "project" in response.json():
        return Project(**response.json()["project"])
    else:
        return None
